"""Viomi Vacuum.

# https://github.com/rytilahti/python-miio/issues/550#issuecomment-552780952
# https://github.com/homebridge-xiaomi-roborock-vacuum/homebridge-xiaomi-roborock-vacuum/blob/ee10cbb3e98dba75d9c97791a6e1fcafc1281591/miio/lib/devices/vacuum.js
# https://github.com/homebridge-xiaomi-roborock-vacuum/homebridge-xiaomi-roborock-vacuum/blob/ee10cbb3e98dba75d9c97791a6e1fcafc1281591/miio/lib/devices/viomivacuum.js

Features:

Main:
- Area/Duration - Missing (get_clean_summary/get_clean_record
- Battery - battery_life
- Dock - set_charge
- Start/Pause - set_mode_withroom
- Modes (Vacuum/Vacuum&Mop/Mop) - set_mop/id_mop
- Fan Speed (Silent/Standard/Medium/Turbo) - set_suction/suction_grade
- Water Level (Low/Medium/High) - set_suction/water_grade

Settings:
- Cleaning history - MISSING (cleanRecord)
- Scheduled cleanup - get_ordertime
- Vacuum along the edges - get_mode/set_mode
- Secondary cleanup - set_repeat/repeat_cleaning
- Mop or vacuum & mod mode - set_moproute/mop_route
- DND(DoNotDisturb) - set_notdisturb/get_notdisturb
- Voice On/Off - set_sound_volume/sound_volume
- Remember Map - remember_map
- Virtual wall/restricted area - MISSING
- Map list - get_maps/rename_map/delete_map/set_map
- Area editor - MISSING
- Reset map - MISSING
- Device leveling - MISSING
- Looking for the vacuum-mop - MISSING (find_me)
- Consumables statistics - get_properties
- Remote Control - MISSING

Misc:
- Get Properties
- Language - set_language
- Led - set_light
- Rooms - get_ordertime (hack)
- Clean History Path - MISSING (historyPath)
- Map plan - MISSING (map_plan)
"""
import itertools
import logging
import time
from collections import defaultdict
from datetime import timedelta
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException
from .utils import pretty_seconds
from .vacuumcontainers import ConsumableStatus, DNDStatus

_LOGGER = logging.getLogger(__name__)


ERROR_CODES = {
    0: "Sleeping and not charging",
    500: "Radar timed out",
    501: "Wheels stuck",
    502: "Low battery",
    503: "Dust bin missing",
    508: "Uneven ground",
    509: "Cliff sensor error",
    510: "Collision sensor error",
    511: "Could not return to dock",
    512: "Could not return to dock",
    513: "Could not navigate",
    514: "Vacuum stuck",
    515: "Charging error",
    516: "Mop temperature error",
    521: "Water tank is not installed",
    522: "Mop is not installed",
    525: "Insufficient water in water tank",
    527: "Remove mop",
    528: "Dust bin missing",
    529: "Mop and water tank missing",
    530: "Mop and water tank missing",
    531: "Water tank is not installed",
    2101: "Unsufficient battery, continuing cleaning after recharge",
    2103: "Charging",
    2105: "Fully charged",
}


class ViomiVacuumException(DeviceException):
    """Exception raised by Viomi Vacuum."""


class ViomiPositionPoint:
    """Vacuum position coordinate."""

    def __init__(self, pos_x, pos_y, phi, update, plan_multiplicator=1):
        self._pos_x = pos_x
        self._pos_y = pos_y
        self.phi = phi
        self.update = update
        self._plan_multiplicator = plan_multiplicator

    @property
    def pos_x(self):
        """X coordinate with multiplicator."""
        return self._pos_x * self._plan_multiplicator

    @property
    def pos_y(self):
        """Y coordinate with multiplicator."""
        return self._pos_y * self._plan_multiplicator

    def image_pos_x(self, offset, img_center):
        """X coordinate on an image."""
        return self.pos_x - offset + img_center

    def image_pos_y(self, offset, img_center):
        """Y coordinate on an image."""
        return self.pos_y - offset + img_center

    def __repr__(self) -> str:
        return "<ViomiPositionPoint x: {}, y: {}, phi: {}, update {}>".format(
            self.pos_x, self.pos_y, self.phi, self.update
        )

    def __eq__(self, value) -> bool:
        return (
            self.pos_x == value.pos_x
            and self.pos_y == value.pos_y
            and self.phi == value.phi
        )


class ViomiConsumableStatus(ConsumableStatus):
    """Consumable container for viomi vacuums.

    Note that this exposes `mop` and `mop_left` that are not available in the base
    class, while returning zeroed timedeltas for `sensor_dirty` and `sensor_dirty_left`
    which it doesn't report.
    """

    def __init__(self, data: List[int]) -> None:
        # [17, 17, 17, 17]
        self.data = {
            "main_brush_work_time": data[0] * 60 * 60,
            "side_brush_work_time": data[1] * 60 * 60,
            "filter_work_time": data[2] * 60 * 60,
            "mop_dirty_time": data[3] * 60 * 60,
        }
        self.side_brush_total = timedelta(hours=180)
        self.main_brush_total = timedelta(hours=360)
        self.filter_total = timedelta(hours=180)
        self.mop_total = timedelta(hours=180)
        self.sensor_dirty_total = timedelta(seconds=0)

    @property
    def mop(self) -> timedelta:
        """Return ``sensor_dirty_time``"""
        return pretty_seconds(self.data["mop_dirty_time"])

    @property
    def mop_left(self) -> timedelta:
        """How long until the mop should be changed."""
        return self.mop_total - self.mop

    @property
    def sensor_dirty(self) -> timedelta:
        """Viomi has no sensor dirty, so we return zero here."""
        return timedelta(seconds=0)

    @property
    def sensor_dirty_left(self) -> timedelta:
        """Viomi has no sensor dirty, so we return zero here."""
        return self.sensor_dirty_total - self.sensor_dirty


class ViomiVacuumSpeed(Enum):
    Silent = 0
    Standard = 1
    Medium = 2
    Turbo = 3


class ViomiVacuumState(Enum):
    Unknown = -1
    IdleNotDocked = 0
    Idle = 1
    Idle2 = 2
    Cleaning = 3
    Returning = 4
    Docked = 5
    VacuumingAndMopping = 6


class ViomiMode(Enum):
    Vacuum = 0  # No Mop, Vacuum only
    VacuumAndMop = 1
    Mop = 2
    CleanZone = 3
    CleanSpot = 4


class ViomiLanguage(Enum):
    CN = 1  # Chinese (default)
    EN = 2  # English


class ViomiLedState(Enum):
    Off = 0
    On = 1


class ViomiCarpetTurbo(Enum):
    Off = 0
    Medium = 1
    Turbo = 2


class ViomiMovementDirection(Enum):
    Forward = 1
    Left = 2  # Rotate
    Right = 3  # Rotate
    Backward = 4
    Stop = 5
    Unknown = 10


class ViomiBinType(Enum):
    Vacuum = 1
    Water = 2
    VacuumAndWater = 3
    NoBin = 0


class ViomiWaterGrade(Enum):
    Low = 11
    Medium = 12
    High = 13


class ViomiRoutePattern(Enum):
    """Mopping pattern."""

    S = 0
    Y = 1


class ViomiEdgeState(Enum):
    Off = 0
    Unknown = 1
    On = 2
    # NOTE: When I got 5, the device was super slow
    # Shutdown and restart device fixed the issue
    Unknown2 = 5


class ViomiVacuumStatus(DeviceStatus):
    def __init__(self, data):
        # ["run_state","mode","err_state","battary_life","box_type","mop_type","s_time","s_area",
        # "suction_grade","water_grade","remember_map","has_map","is_mop","has_newmap"]'
        # 1,               11,           1,            1,         1,       0          ]
        self.data = data

    @property
    def state(self):
        """State of the vacuum."""
        try:
            return ViomiVacuumState(self.data["run_state"])
        except ValueError:
            _LOGGER.warning("Unknown vacuum state: %s", self.data["run_state"])
            return ViomiVacuumState.Unknown

    @property
    def edge_state(self) -> ViomiEdgeState:
        """Vaccum along the edges.

        The settings is valid once
        0: Off
        1: Unknown
        2: On
        5: Unknown
        """
        return ViomiEdgeState(self.data["mode"])

    @property
    def mop_installed(self) -> bool:
        """True if the mop is installed."""
        return bool(self.data["mop_type"])

    @property
    def error_code(self) -> int:
        """Error code from vacuum."""
        return self.data["err_state"]

    @property
    def error(self) -> Optional[str]:
        """String presentation for the error code."""
        if self.error_code is None:
            return None

        return ERROR_CODES.get(self.error_code, f"Unknown error {self.error_code}")

    @property
    def battery(self) -> int:
        """Battery in percentage."""
        return self.data["battary_life"]

    @property
    def bin_type(self) -> ViomiBinType:
        """Type of the inserted bin."""
        return ViomiBinType(self.data["box_type"])

    @property
    def clean_time(self) -> timedelta:
        """Cleaning time."""
        return pretty_seconds(self.data["s_time"] * 60)

    @property
    def clean_area(self) -> float:
        """Cleaned area in square meters."""
        return self.data["s_area"]

    @property
    def fanspeed(self) -> ViomiVacuumSpeed:
        """Current fan speed."""
        return ViomiVacuumSpeed(self.data["suction_grade"])

    @command()
    def fan_speed_presets(self) -> Dict[str, int]:
        """Return dictionary containing supported fanspeeds."""
        return {x.name: x.value for x in list(ViomiVacuumSpeed)}

    @property
    def water_grade(self) -> ViomiWaterGrade:
        """Water grade."""
        return ViomiWaterGrade(self.data["water_grade"])

    @property
    def remember_map(self) -> bool:
        """True to remember the map."""
        return bool(self.data["remember_map"])

    @property
    def has_map(self) -> bool:
        """True if device has map?"""
        return bool(self.data["has_map"])

    @property
    def has_new_map(self) -> bool:
        """True if the device has scanned a new map (like a new floor)."""
        return bool(self.data["has_newmap"])

    @property
    def mop_mode(self) -> ViomiMode:
        """Whether mopping is enabled and if so which mode."""
        return ViomiMode(self.data["is_mop"])

    @property
    def current_map_id(self) -> float:
        """Current map id."""
        return self.data["cur_mapid"]

    @property
    def hw_info(self) -> str:
        """Hardware info."""
        return self.data["hw_info"]

    @property
    def charging(self) -> bool:
        """True if battery is charging.

        Note: When the battery is at 100%, device reports that it is not charging.
        """
        return not bool(self.data["is_charge"])

    @property
    def is_on(self) -> bool:
        """True if device is working."""
        return not bool(self.data["is_work"])

    @property
    def light_state(self) -> bool:
        """Led state.

        This seems doing nothing on STYJ02YM
        """
        return bool(self.data["light_state"])

    @property
    def map_number(self) -> int:
        """Number of saved maps."""
        return self.data["map_num"]

    @property
    def mop_route(self) -> ViomiRoutePattern:
        """Pattern mode."""
        return ViomiRoutePattern(self.data["mop_route"])

    # @property
    # def order_time(self) -> int:
    #    """FIXME: ??? int or bool."""
    #    return self.data["order_time"]

    @property
    def repeat_cleaning(self) -> bool:
        """Secondary clean up state.

        True if the cleaning is performed twice
        """
        return self.data["repeat_state"]

    # @property
    # def start_time(self) -> int:
    #    """FIXME: ??? int or bool."""
    #    return self.data["start_time"]

    @property
    def sound_volume(self) -> int:
        """Voice volume level (from 0 to 100%, 0 means Off)."""
        return self.data["v_state"]

    # @property
    # def water_percent(self) -> int:
    #    """FIXME: ??? int or bool."""
    #    return self.data["water_percent"]

    # @property
    # def zone_data(self) -> int:
    #    """FIXME: ??? int or bool."""
    #    return self.data["zone_data"]


def _get_rooms_from_schedules(schedules: List[str]) -> Tuple[bool, Dict]:
    """Read the result of "get_ordertime" command to extract room names and ids.

    The `schedules` input needs to follow the following format
    * ['1_0_32_0_0_0_1_1_11_0_1594139992_2_11_room1_13_room2', ...]
    * [Id_Enabled_Repeatdays_Hour_Minute_?_? _?_?_?_?_NbOfRooms_RoomId_RoomName_RoomId_RoomName_..., ...]

    The function parse get_ordertime output to find room names and ids
    To use this function you need:
    1. to create a scheduled cleanup with the following properties:
    * Hour: 00
    * Minute: 00
    * Select all (minus one) the rooms one by one
    * Set as inactive scheduled cleanup
    2. then to create an other scheduled cleanup with the room missed at
       previous step with the following properties:
    * Hour: 00
    * Minute: 00
    * Select only the missed room
    * Set as inactive scheduled cleanup

    More information:
    * https://github.com/homebridge-xiaomi-roborock-vacuum/homebridge-xiaomi-roborock-vacuum/blob/d73925c0106984a995d290e91a5ba4fcfe0b6444/index.js#L969
    * https://github.com/homebridge-xiaomi-roborock-vacuum/homebridge-xiaomi-roborock-vacuum#semi-automatic
    """
    rooms = {}
    scheduled_found = False
    for raw_schedule in schedules:
        schedule = raw_schedule.split("_")
        # Scheduled cleanup needs to be scheduled for 00:00 and inactive
        if schedule[1] == "0" and schedule[3] == "0" and schedule[4] == "0":
            scheduled_found = True
            raw_rooms = schedule[12:]
            rooms_iter = iter(raw_rooms)
            rooms.update(
                dict(itertools.zip_longest(rooms_iter, rooms_iter, fillvalue=None))
            )
    return scheduled_found, rooms


class ViomiVacuum(Device):
    """Interface for Viomi vacuums (viomi.vacuum.v7)."""

    timeout = 5
    retry_count = 10

    def __init__(
        self, ip: str, token: str = None, start_id: int = 0, debug: int = 0
    ) -> None:
        super().__init__(ip, token, start_id, debug)
        self.manual_seqnum = -1
        self._cache: Dict[str, Any] = {"edge_state": None, "rooms": {}, "maps": {}}

    @command(
        default_output=format_output(
            "\n",
            "General\n"
            "=======\n\n"
            "Hardware version: {result.hw_info}\n"
            "State: {result.state}\n"
            "Working: {result.is_on}\n"
            "Battery status: {result.error}\n"
            "Battery: {result.battery}\n"
            "Charging: {result.charging}\n"
            "Box type: {result.bin_type}\n"
            "Fan speed: {result.fanspeed}\n"
            "Water grade: {result.water_grade}\n"
            "Mop mode: {result.mop_mode}\n"
            "Mop installed: {result.mop_installed}\n"
            "Vacuum along the edges: {result.edge_state}\n"
            "Mop route pattern: {result.mop_route}\n"
            "Secondary Cleanup: {result.repeat_cleaning}\n"
            "Sound Volume: {result.sound_volume}\n"
            "Clean time: {result.clean_time}\n"
            "Clean area: {result.clean_area} m²\n"
            "\n"
            "Map\n"
            "===\n\n"
            "Current map ID: {result.current_map_id}\n"
            "Remember map: {result.remember_map}\n"
            "Has map: {result.has_map}\n"
            "Has new map: {result.has_new_map}\n"
            "Number of maps: {result.map_number}\n"
            "\n"
            "Unknown properties\n"
            "=================\n\n"
            "Light state: {result.light_state}\n"
            # "Order time: {result.order_time}\n"
            # "Start time: {result.start_time}\n"
            # "water_percent: {result.water_percent}\n"
            # "zone_data: {result.zone_data}\n",
        )
    )
    def status(self) -> ViomiVacuumStatus:
        """Retrieve properties."""
        properties = [
            "battary_life",
            "box_type",
            "cur_mapid",
            "err_state",
            "has_map",
            "has_newmap",
            "hw_info",
            "is_charge",
            "is_mop",
            "is_work",
            "light_state",
            "map_num",
            "mode",
            "mop_route",
            "mop_type",
            "remember_map",
            "repeat_state",
            "run_state",
            "s_area",
            "s_time",
            "suction_grade",
            "v_state",
            "water_grade",
            # The following list of properties existing but
            # there are not used in the code
            # "order_time",
            # "start_time",
            # "water_percent",
            # "zone_data",
            # "sw_info",
            # "main_brush_hours",
            # "main_brush_life",
            # "side_brush_hours",
            # "side_brush_life",
            # "mop_hours",
            # "mop_life",
            # "hypa_hours",
            # "hypa_life",
        ]

        values = self.get_properties(properties)

        return ViomiVacuumStatus(defaultdict(lambda: None, zip(properties, values)))

    @command()
    def home(self):
        """Return to home."""
        self.send("set_charge", [1])

    @command()
    def start(self):
        """Start cleaning."""
        # params: [edge, 1, roomIds.length, *list_of_room_ids]
        # - edge: see ViomiEdgeState
        # - 1: start cleaning (2 pause, 0 stop)
        # - roomIds.length
        # - *room_id_list
        # 3rd param of set_mode_withroom is room_array_len and next are
        # room ids ([0, 1, 3, 11, 12, 13] = start cleaning rooms 11-13).
        # room ids are encoded in map and it's part of cloud api so best way
        # to get it is log between device <> mi home app
        # (before map format is supported).
        self._cache["edge_state"] = self.get_properties(["mode"])
        self.send("set_mode_withroom", self._cache["edge_state"] + [1, 0])

    @command(
        click.option(
            "--rooms",
            "-r",
            multiple=True,
            help="Rooms name or room id. Can be used multiple times",
        )
    )
    def start_with_room(self, rooms):
        """Start cleaning specific rooms."""
        if not self._cache["rooms"]:
            self.get_rooms()
        reverse_rooms = {v: k for k, v in self._cache["rooms"].items()}
        room_ids = []
        for room in rooms:
            if room in self._cache["rooms"]:
                room_ids.append(int(room))
            elif room in reverse_rooms:
                room_ids.append(int(reverse_rooms[room]))
            else:
                room_keys = ", ".join(self._cache["rooms"].keys())
                room_ids = ", ".join(self._cache["rooms"].values())
                raise DeviceException(
                    f"Room {room} is unknown, it must be in {room_keys} or {room_ids}"
                )

        self._cache["edge_state"] = self.get_properties(["mode"])
        self.send(
            "set_mode_withroom",
            self._cache["edge_state"] + [1, len(room_ids)] + room_ids,
        )

    @command()
    def pause(self):
        """Pause cleaning."""
        # params: [edge_state, 0]
        # - edge: see ViomiEdgeState
        # - 2: pause cleaning
        if not self._cache["edge_state"]:
            self._cache["edge_state"] = self.get_properties(["mode"])
        self.send("set_mode", self._cache["edge_state"] + [2])

    @command()
    def stop(self):
        """Validate that Stop cleaning."""
        # params: [edge_state, 0]
        # - edge: see ViomiEdgeState
        # - 0: stop cleaning
        if not self._cache["edge_state"]:
            self._cache["edge_state"] = self.get_properties(["mode"])
        self.send("set_mode", self._cache["edge_state"] + [0])

    @command(click.argument("mode", type=EnumType(ViomiMode)))
    def clean_mode(self, mode: ViomiMode):
        """Set the cleaning mode.

        [vacuum, vacuumAndMop, mop, cleanzone, cleanspot]
        """
        self.send("set_mop", [mode.value])

    @command(click.argument("speed", type=EnumType(ViomiVacuumSpeed)))
    def set_fan_speed(self, speed: ViomiVacuumSpeed):
        """Set fanspeed [silent, standard, medium, turbo]."""
        self.send("set_suction", [speed.value])

    @command(click.argument("watergrade", type=EnumType(ViomiWaterGrade)))
    def set_water_grade(self, watergrade: ViomiWaterGrade):
        """Set water grade.

        [low, medium, high]
        """
        self.send("set_suction", [watergrade.value])

    def get_positions(self, plan_multiplicator=1) -> List[ViomiPositionPoint]:
        """Return the last positions.

        plan_multiplicator scale up the coordinates values
        """
        results = self.send("get_curpos", [])
        positions = []
        # Group result 4 by 4
        for res in [i for i in zip(*(results[i::4] for i in range(4)))]:
            # ignore type require for mypy error
            # "ViomiPositionPoint" gets multiple values for keyword argument "plan_multiplicator"
            positions.append(
                ViomiPositionPoint(*res, plan_multiplicator=plan_multiplicator)  # type: ignore
            )
        return positions

    @command()
    def get_current_position(self) -> Optional[ViomiPositionPoint]:
        """Return the current position."""
        positions = self.get_positions()
        if positions:
            return positions[-1]
        return None

    # MISSING cleaning history

    @command()
    def get_scheduled_cleanup(self):
        """Not implemented yet."""
        # Needs to reads and understand the return of:
        # self.send("get_ordertime", [])
        # [id, enabled, repeatdays, hour, minute, ?, ? , ?, ?, ?, ?, nb_of_rooms, room_id, room_name, room_id, room_name, ...]
        raise NotImplementedError()

    @command()
    def add_timer(self):
        """Not implemented yet."""
        # Needs to reads and understand:
        # self.send("set_ordertime", [????])
        raise NotImplementedError()

    @command()
    def delete_timer(self):
        """Not implemented yet."""
        # Needs to reads and understand:
        # self.send("det_ordertime", [shedule_id])
        raise NotImplementedError()

    @command(click.argument("state", type=EnumType(ViomiEdgeState)))
    def set_edge(self, state: ViomiEdgeState):
        """Vacuum along edges.

        This is valid for a single cleaning.
        """
        return self.send("set_mode", [state.value])

    @command(click.argument("state", type=bool))
    def set_repeat(self, state: bool):
        """Set or Unset repeat mode (Secondary cleanup)."""
        return self.send("set_repeat", [int(state)])

    @command(click.argument("mop_mode", type=EnumType(ViomiRoutePattern)))
    def set_route_pattern(self, mop_mode: ViomiRoutePattern):
        """Set the mop route pattern."""
        self.send("set_moproute", [mop_mode.value])

    @command()
    def dnd_status(self):
        """Returns do-not-disturb status."""
        status = self.send("get_notdisturb")
        return DNDStatus(
            dict(
                enabled=status[0],
                start_hour=status[1],
                start_minute=status[2],
                end_hour=status[3],
                end_minute=status[4],
            )
        )

    @command(
        click.option("--disable", is_flag=True),
        click.argument("start_hr", type=int),
        click.argument("start_min", type=int),
        click.argument("end_hr", type=int),
        click.argument("end_min", type=int),
    )
    def set_dnd(
        self, disable: bool, start_hr: int, start_min: int, end_hr: int, end_min: int
    ):
        """Set do-not-disturb.

        :param int start_hr: Start hour
        :param int start_min: Start minute
        :param int end_hr: End hour
        :param int end_min: End minute
        """
        return self.send(
            "set_notdisturb",
            [0 if disable else 1, start_hr, start_min, end_hr, end_min],
        )

    @command(click.argument("volume", type=click.IntRange(0, 10)))
    def set_sound_volume(self, volume: int):
        """Switch the voice on or off."""
        enabled = 1
        if volume == 0:
            enabled = 0
        return self.send("set_voice", [enabled, volume])

    @command(click.argument("state", type=bool))
    def set_remember_map(self, state: bool):
        """Set remember map state."""
        return self.send("set_remember", [int(state)])

    # MISSING: Virtual wall/restricted area

    @command()
    def get_maps(self) -> List[Dict[str, Any]]:
        """Return map list.

        [{'name': 'MapName1', 'id': 1598622255, 'cur': False},
         {'name': 'MapName2', 'id': 1599508355, 'cur': True},
          ...]
        """
        if not self._cache["maps"]:
            self._cache["maps"] = self.send("get_map")
        return self._cache["maps"]

    @command(click.argument("map_id", type=int))
    def set_map(self, map_id: int):
        """Change current map."""
        maps = self.get_maps()
        if map_id not in [m["id"] for m in maps]:
            raise ViomiVacuumException("Map id {} doesn't exists".format(map_id))
        return self.send("set_map", [map_id])

    @command(click.argument("map_id", type=int))
    def delete_map(self, map_id: int):
        """Delete map."""
        maps = self.get_maps()
        if map_id not in [m["id"] for m in maps]:
            raise ViomiVacuumException("Map id {} doesn't exists".format(map_id))
        return self.send("del_map", [map_id])

    @command(
        click.argument("map_id", type=int),
        click.argument("map_name", type=str),
    )
    def rename_map(self, map_id: int, map_name: str):
        """Rename map."""
        maps = self.get_maps()
        if map_id not in [m["id"] for m in maps]:
            raise ViomiVacuumException(f"Map id {map_id} doesn't exists")
        return self.send("rename_map", {"mapID": map_id, "name": map_name})

    @command(
        click.option("--map-id", type=int, default=None),
        click.option("--map-name", type=str, default=None),
        click.option("--refresh", type=bool, default=False),
    )
    def get_rooms(
        self, map_id: int = None, map_name: str = None, refresh: bool = False
    ):
        """Return room ids and names."""
        if self._cache["rooms"] and not refresh:
            return self._cache["rooms"]

        # TODO: map_name and map_id are just dead code here?
        if map_name:
            maps = self.get_maps()
            map_ids = [map_["id"] for map_ in maps if map_["name"] == map_name]
            if not map_ids:
                map_names = ", ".join([m["name"] for m in maps])
                raise ViomiVacuumException(
                    f"Error: Bad map name, should be in {map_names}"
                )
        elif map_id:
            maps = self.get_maps()
            if map_id not in [m["id"] for m in maps]:
                map_ids_str = ", ".join([str(m["id"]) for m in maps])
                raise ViomiVacuumException(
                    f"Error: Bad map id, should be in {map_ids_str}"
                )
        # Get scheduled cleanup
        schedules = self.send("get_ordertime", [])
        scheduled_found, rooms = _get_rooms_from_schedules(schedules)
        if not scheduled_found:
            msg = (
                "Fake schedule not found. "
                "Please create a scheduled cleanup with the "
                "following properties:\n"
                "* Hour: 00\n"
                "* Minute: 00\n"
                "* Select all (minus one) the rooms one by one\n"
                "* Set as inactive scheduled cleanup\n"
                "Then create a scheduled cleanup with the room missed at "
                "previous step with the following properties:\n"
                "* Hour: 00\n"
                "* Minute: 00\n"
                "* Select only the missed room\n"
                "* Set as inactive scheduled cleanup\n"
            )
            raise ViomiVacuumException(msg)

        self._cache["rooms"] = rooms
        return rooms

    # MISSING Area editor

    # MISSING Reset map

    # MISSING Device leveling

    # MISSING Looking for the vacuum-mop

    @command()
    def consumable_status(self) -> ViomiConsumableStatus:
        """Return information about consumables."""
        return ViomiConsumableStatus(self.send("get_consumables"))

    @command(
        click.argument("direction", type=EnumType(ViomiMovementDirection)),
        click.option(
            "--duration",
            type=float,
            default=0.5,
            help="number of seconds to perform this movement",
        ),
    )
    def move(self, direction: ViomiMovementDirection, duration=0.5):
        """Manual movement."""
        start = time.time()
        while time.time() - start < duration:
            self.send("set_direction", [direction.value])
            time.sleep(0.1)
        self.send("set_direction", [ViomiMovementDirection.Stop.value])

    @command(click.argument("language", type=EnumType(ViomiLanguage)))
    def set_language(self, language: ViomiLanguage):
        """Set the device's audio language.

        This seems doing nothing on STYJ02YM
        """
        return self.send("set_language", [language.value])

    @command(click.argument("state", type=EnumType(ViomiLedState)))
    def led(self, state: ViomiLedState):
        """Switch the button leds on or off.

        This seems doing nothing on STYJ02YM
        """
        return self.send("set_light", [state.value])

    @command(click.argument("mode", type=EnumType(ViomiCarpetTurbo)))
    def carpet_mode(self, mode: ViomiCarpetTurbo):
        """Set the carpet mode.

        This seems doing nothing on STYJ02YM
        """
        return self.send("set_carpetturbo", [mode.value])



# flake8: noqa
try:
    # python 3.7 and earlier
    from importlib_metadata import version  # type: ignore
except ImportError:
    # python 3.8 and later
    from importlib.metadata import version  # type: ignore

from miio.airconditioner_miot import AirConditionerMiot
from miio.airconditioningcompanion import (
    AirConditioningCompanion,
    AirConditioningCompanionV3,
)
from miio.airconditioningcompanionMCN import AirConditioningCompanionMcn02
from miio.airdehumidifier import AirDehumidifier
from miio.airfresh import AirFresh, AirFreshVA4
from miio.airfresh_t2017 import AirFreshA1, AirFreshT2017
from miio.airhumidifier import AirHumidifier, AirHumidifierCA1, AirHumidifierCB1
from miio.airhumidifier_jsq import AirHumidifierJsq
from miio.airhumidifier_miot import AirHumidifierMiot
from miio.airhumidifier_mjjsq import AirHumidifierMjjsq
from miio.airpurifier import AirPurifier
from miio.airpurifier_airdog import AirDogX3, AirDogX5, AirDogX7SM
from miio.airpurifier_miot import AirPurifierMB4, AirPurifierMiot
from miio.airqualitymonitor import AirQualityMonitor
from miio.airqualitymonitor_miot import AirQualityMonitorCGDN1
from miio.aqaracamera import AqaraCamera
from miio.ceil import Ceil
from miio.chuangmi_camera import ChuangmiCamera
from miio.chuangmi_ir import ChuangmiIr
from miio.chuangmi_plug import ChuangmiPlug, Plug, PlugV1, PlugV3
from miio.cooker import Cooker
from miio.curtain_youpin import CurtainMiot
from miio.device import Device, DeviceStatus
from miio.dreamevacuum_miot import DreameVacuumMiot
from miio.exceptions import DeviceError, DeviceException
from miio.fan import Fan, FanP5, FanSA1, FanV2, FanZA1, FanZA4
from miio.fan_leshow import FanLeshow
from miio.fan_miot import Fan1C, FanMiot, FanP9, FanP10, FanP11
from miio.gateway import Gateway
from miio.heater import Heater
from miio.heater_miot import HeaterMiot
from miio.huizuo import Huizuo, HuizuoLampFan, HuizuoLampHeater, HuizuoLampScene
from miio.miot_device import MiotDevice
from miio.philips_bulb import PhilipsBulb, PhilipsWhiteBulb
from miio.philips_eyecare import PhilipsEyecare
from miio.philips_moonlight import PhilipsMoonlight
from miio.philips_rwread import PhilipsRwread
from miio.powerstrip import PowerStrip
from miio.protocol import Message, Utils
from miio.pwzn_relay import PwznRelay
from miio.roidmivacuum_miot import RoidmiVacuumMiot
from miio.scishare_coffeemaker import ScishareCoffee
from miio.toiletlid import Toiletlid
from miio.vacuum import Vacuum, VacuumException
from miio.vacuum_tui import VacuumTUI
from miio.vacuumcontainers import (
    CleaningDetails,
    CleaningSummary,
    ConsumableStatus,
    DNDStatus,
    Timer,
    VacuumStatus,
)
from miio.viomivacuum import ViomiVacuum
from miio.walkingpad import Walkingpad
from miio.waterpurifier import WaterPurifier
from miio.waterpurifier_yunmi import WaterPurifierYunmi
from miio.wifirepeater import WifiRepeater
from miio.wifispeaker import WifiSpeaker
from miio.yeelight import Yeelight
from miio.yeelight_dual_switch import YeelightDualControlModule

from miio.discovery import Discovery

__version__ = version("python-miio")



import logging
import sys

import click

import miio  # noqa: E402
from miio.click_common import ExceptionHandlerGroup, validate_ip, validate_token
from miio.miioprotocol import MiIOProtocol

_LOGGER = logging.getLogger(__name__)
pass_dev = click.make_pass_decorator(miio.Ceil)


def validate_percentage(ctx, param, value):
    value = int(value)
    if value < 1 or value > 100:
        raise click.BadParameter("Should be a positive int between 1-100.")
    return value


def validate_seconds(ctx, param, value):
    value = int(value)
    if value < 0 or value > 21600:
        raise click.BadParameter("Should be a positive int between 1-21600.")
    return value


def validate_scene(ctx, param, value):
    value = int(value)
    if value < 1 or value > 4:
        raise click.BadParameter("Should be a positive int between 1-4.")
    return value


@click.group(invoke_without_command=True, cls=ExceptionHandlerGroup)
@click.option("--ip", envvar="DEVICE_IP", callback=validate_ip)
@click.option("--token", envvar="DEVICE_TOKEN", callback=validate_token)
@click.option("-d", "--debug", default=False, count=True)
@click.pass_context
def cli(ctx, ip: str, token: str, debug: int):
    """A tool to command Xiaomi Philips LED Ceiling Lamp."""

    if debug:
        logging.basicConfig(level=logging.DEBUG)
        _LOGGER.info("Debug mode active")
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.warning(
        "This script is deprecated and will be removed soon, use `miiocli ceil` instead"
    )

    # if we are scanning, we do not try to connect.
    if ctx.invoked_subcommand == "discover":
        return

    if ip is None or token is None:
        click.echo("You have to give ip and token!")
        sys.exit(-1)

    dev = miio.Ceil(ip, token, debug)
    _LOGGER.debug("Connecting to %s with token %s", ip, token)

    ctx.obj = dev

    if ctx.invoked_subcommand is None:
        ctx.invoke(status)


@cli.command()
def discover():
    """Search for plugs in the network."""
    MiIOProtocol.discover()


@cli.command()
@pass_dev
def status(dev: miio.Ceil):
    """Returns the state information."""
    res = dev.status()
    if not res:
        return  # bail out

    click.echo(click.style("Power: %s" % res.power, bold=True))
    click.echo("Brightness: %s" % res.brightness)
    click.echo("Color temperature: %s" % res.color_temperature)
    click.echo("Scene: %s" % res.scene)
    click.echo("Smart Night Light: %s" % res.smart_night_light)
    click.echo("Auto CCT: %s" % res.automatic_color_temperature)
    click.echo(
        "Countdown of the delayed turn off: %s seconds" % res.delay_off_countdown
    )


@cli.command()
@pass_dev
def on(dev: miio.Ceil):
    """Power on."""
    click.echo("Power on: %s" % dev.on())


@cli.command()
@pass_dev
def off(dev: miio.Ceil):
    """Power off."""
    click.echo("Power off: %s" % dev.off())


@cli.command()
@click.argument("level", callback=validate_percentage, required=True)
@pass_dev
def set_brightness(dev: miio.Ceil, level):
    """Set brightness level."""
    click.echo("Brightness: %s" % dev.set_brightness(level))


@cli.command()
@click.argument("level", callback=validate_percentage, required=True)
@pass_dev
def set_color_temperature(dev: miio.Ceil, level):
    """Set CCT level."""
    click.echo("Color temperature level: %s" % dev.set_color_temperature(level))


@cli.command()
@click.argument("seconds", callback=validate_seconds, required=True)
@pass_dev
def delay_off(dev: miio.Ceil, seconds):
    """Set delay off in seconds."""
    click.echo("Delay off: %s" % dev.delay_off(seconds))


@cli.command()
@click.argument("scene", callback=validate_scene, required=True)
@pass_dev
def set_scene(dev: miio.Ceil, scene):
    """Set scene number."""
    click.echo("Eyecare Scene: %s" % dev.set_scene(scene))


@cli.command()
@pass_dev
def smart_night_light_on(dev: miio.Ceil):
    """Smart Night Light on."""
    click.echo("Smart Night Light On: %s" % dev.smart_night_light_on())


@cli.command()
@pass_dev
def smart_night_light_off(dev: miio.Ceil):
    """Smart Night Light off."""
    click.echo("Smart Night Light Off: %s" % dev.smart_night_light_off())


@cli.command()
@pass_dev
def automatic_color_temperature_on(dev: miio.Ceil):
    """Auto CCT on."""
    click.echo("Auto CCT On: %s" % dev.automatic_color_temperature_on())


@cli.command()
@pass_dev
def automatic_color_temperature_off(dev: miio.Ceil):
    """Auto CCT on."""
    click.echo("Auto CCT Off: %s" % dev.automatic_color_temperature_off())


if __name__ == "__main__":
    cli()



import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .click_common import command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException
from .utils import deprecated

_LOGGER = logging.getLogger(__name__)

MODEL_CHUANGMI_PLUG_V3 = "chuangmi.plug.v3"
MODEL_CHUANGMI_PLUG_V1 = "chuangmi.plug.v1"
MODEL_CHUANGMI_PLUG_M1 = "chuangmi.plug.m1"
MODEL_CHUANGMI_PLUG_M3 = "chuangmi.plug.m3"
MODEL_CHUANGMI_PLUG_V2 = "chuangmi.plug.v2"
MODEL_CHUANGMI_PLUG_HMI205 = "chuangmi.plug.hmi205"
MODEL_CHUANGMI_PLUG_HMI206 = "chuangmi.plug.hmi206"
MODEL_CHUANGMI_PLUG_HMI208 = "chuangmi.plug.hmi208"

AVAILABLE_PROPERTIES = {
    MODEL_CHUANGMI_PLUG_V1: ["on", "usb_on", "temperature"],
    MODEL_CHUANGMI_PLUG_V3: ["on", "usb_on", "temperature", "wifi_led"],
    MODEL_CHUANGMI_PLUG_M1: ["power", "temperature"],
    MODEL_CHUANGMI_PLUG_M3: ["power", "temperature"],
    MODEL_CHUANGMI_PLUG_V2: ["power", "temperature"],
    MODEL_CHUANGMI_PLUG_HMI205: ["power", "temperature"],
    MODEL_CHUANGMI_PLUG_HMI206: ["power", "temperature"],
    MODEL_CHUANGMI_PLUG_HMI208: ["power", "usb_on", "temperature"],
}


class ChuangmiPlugStatus(DeviceStatus):
    """Container for status reports from the plug."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Response of a Chuangmi Plug V1 (chuangmi.plug.v1)

        { 'power': True, 'usb_on': True, 'temperature': 32 }

        Response of a Chuangmi Plug V3 (chuangmi.plug.v3):
        { 'on': True, 'usb_on': True, 'temperature': 32, 'wifi_led': True }
        """
        self.data = data

    @property
    def power(self) -> bool:
        """Current power state."""
        if "on" in self.data:
            return self.data["on"] is True or self.data["on"] == "on"
        elif "power" in self.data:
            return self.data["power"] == "on"

        raise DeviceException("There was neither 'on' or 'power' in data")

    @property
    def is_on(self) -> bool:
        """True if device is on."""
        return self.power

    @property
    def temperature(self) -> int:
        return self.data["temperature"]

    @property
    def usb_power(self) -> Optional[bool]:
        """True if USB is on."""
        if "usb_on" in self.data and self.data["usb_on"] is not None:
            return self.data["usb_on"]
        return None

    @property
    def load_power(self) -> Optional[float]:
        """Current power load, if available."""
        if "load_power" in self.data and self.data["load_power"] is not None:
            return float(self.data["load_power"])
        return None

    @property
    def wifi_led(self) -> Optional[bool]:
        """True if the wifi led is turned on."""
        if "wifi_led" in self.data and self.data["wifi_led"] is not None:
            return self.data["wifi_led"] == "on"
        return None


class ChuangmiPlug(Device):
    """Main class representing the Chuangmi Plug."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_CHUANGMI_PLUG_M1,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_CHUANGMI_PLUG_M1

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "USB Power: {result.usb_power}\n"
            "Temperature: {result.temperature} °C\n"
            "Load power: {result.load_power}\n"
            "WiFi LED: {result.wifi_led}\n",
        )
    )
    def status(self) -> ChuangmiPlugStatus:
        """Retrieve properties."""
        properties = AVAILABLE_PROPERTIES[self.model].copy()
        values = self.get_properties(properties)

        if self.model == MODEL_CHUANGMI_PLUG_V3:
            load_power = self.send("get_power")  # Response: [300]
            if len(load_power) == 1:
                properties.append("load_power")
                values.append(load_power[0] * 0.01)

        return ChuangmiPlugStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        if self.model == MODEL_CHUANGMI_PLUG_V1:
            return self.send("set_on")

        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        if self.model == MODEL_CHUANGMI_PLUG_V1:
            return self.send("set_off")

        return self.send("set_power", ["off"])

    @command(default_output=format_output("Powering USB on"))
    def usb_on(self):
        """Power on."""
        return self.send("set_usb_on")

    @command(default_output=format_output("Powering USB off"))
    def usb_off(self):
        """Power off."""
        return self.send("set_usb_off")

    @command(
        click.argument("wifi_led", type=bool),
        default_output=format_output(
            lambda wifi_led: "Turning on WiFi LED"
            if wifi_led
            else "Turning off WiFi LED"
        ),
    )
    def set_wifi_led(self, wifi_led: bool):
        """Set the wifi led on/off."""
        if wifi_led:
            return self.send("set_wifi_led", ["on"])
        else:
            return self.send("set_wifi_led", ["off"])


@deprecated(
    "This device class is deprecated. Please use the ChuangmiPlug "
    "class in future and select a model by parameter 'model'."
)
class Plug(ChuangmiPlug):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
    ) -> None:
        super().__init__(
            ip, token, start_id, debug, lazy_discover, model=MODEL_CHUANGMI_PLUG_M1
        )


@deprecated(
    "This device class is deprecated. Please use the ChuangmiPlug "
    "class in future and select a model by parameter 'model'."
)
class PlugV1(ChuangmiPlug):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
    ) -> None:
        super().__init__(
            ip, token, start_id, debug, lazy_discover, model=MODEL_CHUANGMI_PLUG_V1
        )


@deprecated(
    "This device class is deprecated. Please use the ChuangmiPlug "
    "class in future and select a model by parameter 'model'."
)
class PlugV3(ChuangmiPlug):
    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
    ) -> None:
        super().__init__(
            ip, token, start_id, debug, lazy_discover, model=MODEL_CHUANGMI_PLUG_V3
        )



import enum
import logging
from collections import defaultdict
from typing import Any, Dict

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_PHILIPS_LIGHT_RWREAD = "philips.light.rwread"

AVAILABLE_PROPERTIES = {
    MODEL_PHILIPS_LIGHT_RWREAD: ["power", "bright", "dv", "snm", "flm", "chl", "flmv"]
}


class PhilipsRwreadException(DeviceException):
    pass


class MotionDetectionSensitivity(enum.Enum):
    Low = 1
    Medium = 2
    High = 3


class PhilipsRwreadStatus(DeviceStatus):
    """Container for status reports from Xiaomi Philips RW Read."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Response of a RW Read (philips.light.rwread):

        {'power': 'on', 'bright': 53, 'dv': 0, 'snm': 1,
         'flm': 0, 'chl': 0, 'flmv': 0}
        """
        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return self.data["power"]

    @property
    def is_on(self) -> bool:
        """True if the device is turned on."""
        return self.power == "on"

    @property
    def brightness(self) -> int:
        """Current brightness."""
        return self.data["bright"]

    @property
    def delay_off_countdown(self) -> int:
        """Countdown until turning off in seconds."""
        return self.data["dv"]

    @property
    def scene(self) -> int:
        """Current fixed scene."""
        return self.data["snm"]

    @property
    def motion_detection(self) -> bool:
        """True if motion detection is enabled."""
        return self.data["flm"] == 1

    @property
    def motion_detection_sensitivity(self) -> MotionDetectionSensitivity:
        """The sensitivity of the motion detection."""
        return MotionDetectionSensitivity(self.data["flmv"])

    @property
    def child_lock(self) -> bool:
        """True if child lock is enabled."""
        return self.data["chl"] == 1


class PhilipsRwread(Device):
    """Main class representing Xiaomi Philips RW Read."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_PHILIPS_LIGHT_RWREAD,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_PHILIPS_LIGHT_RWREAD

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Brightness: {result.brightness}\n"
            "Delayed turn off: {result.delay_off_countdown}\n"
            "Scene: {result.scene}\n"
            "Motion detection: {result.motion_detection}\n"
            "Motion detection sensitivity: {result.motion_detection_sensitivity}\n"
            "Child lock: {result.child_lock}\n",
        )
    )
    def status(self) -> PhilipsRwreadStatus:
        """Retrieve properties."""
        properties = AVAILABLE_PROPERTIES[self.model]
        values = self.get_properties(properties)

        return PhilipsRwreadStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("level", type=int),
        default_output=format_output("Setting brightness to {level}"),
    )
    def set_brightness(self, level: int):
        """Set brightness level of the primary light."""
        if level < 1 or level > 100:
            raise PhilipsRwreadException("Invalid brightness: %s" % level)

        return self.send("set_bright", [level])

    @command(
        click.argument("number", type=int),
        default_output=format_output("Setting fixed scene to {number}"),
    )
    def set_scene(self, number: int):
        """Set one of the fixed eyecare user scenes."""
        if number < 1 or number > 4:
            raise PhilipsRwreadException("Invalid fixed scene number: %s" % number)

        return self.send("apply_fixed_scene", [number])

    @command(
        click.argument("seconds", type=int),
        default_output=format_output("Setting delayed turn off to {seconds} seconds"),
    )
    def delay_off(self, seconds: int):
        """Set delay off in seconds."""

        if seconds < 0:
            raise PhilipsRwreadException(
                "Invalid value for a delayed turn off: %s" % seconds
            )

        return self.send("delay_off", [seconds])

    @command(
        click.argument("motion_detection", type=bool),
        default_output=format_output(
            lambda motion_detection: "Turning on motion detection"
            if motion_detection
            else "Turning off motion detection"
        ),
    )
    def set_motion_detection(self, motion_detection: bool):
        """Set motion detection on/off."""
        return self.send("enable_flm", [int(motion_detection)])

    @command(
        click.argument("sensitivity", type=EnumType(MotionDetectionSensitivity)),
        default_output=format_output(
            "Setting motion detection sensitivity to {sensitivity}"
        ),
    )
    def set_motion_detection_sensitivity(self, sensitivity: MotionDetectionSensitivity):
        """Set motion detection sensitivity."""
        return self.send("set_flmvalue", [sensitivity.value])

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        return self.send("enable_chl", [int(lock)])



from unittest import TestCase

import pytest

from miio import Huizuo, HuizuoLampFan, HuizuoLampHeater
from miio.huizuo import MODEL_HUIZUO_FANWY  # Fan model extended
from miio.huizuo import MODEL_HUIZUO_FANWY2  # Fan model basic
from miio.huizuo import MODEL_HUIZUO_PIS123  # Basic model
from miio.huizuo import MODEL_HUIZUO_WYHEAT  # Heater model
from miio.huizuo import HuizuoException

from .dummies import DummyMiotDevice

_INITIAL_STATE = {
    "power": True,
    "brightness": 60,
    "color_temp": 4000,
}

_INITIAL_STATE_FAN = {
    "power": True,
    "brightness": 60,
    "color_temp": 4000,
    "fan_power": False,
    "fan_level": 60,
    "fan_motor_reverse": True,
    "fan_mode": 1,
}

_INITIAL_STATE_HEATER = {
    "power": True,
    "brightness": 60,
    "color_temp": 4000,
    "heater_power": True,
    "heat_level": 2,
}


class DummyHuizuo(DummyMiotDevice, Huizuo):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE
        self.model = MODEL_HUIZUO_PIS123
        super().__init__(*args, **kwargs)


class DummyHuizuoFan(DummyMiotDevice, HuizuoLampFan):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE_FAN
        self.model = MODEL_HUIZUO_FANWY
        super().__init__(*args, **kwargs)


class DummyHuizuoFan2(DummyMiotDevice, HuizuoLampFan):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE_FAN
        self.model = MODEL_HUIZUO_FANWY2
        super().__init__(*args, **kwargs)


class DummyHuizuoHeater(DummyMiotDevice, HuizuoLampHeater):
    def __init__(self, *args, **kwargs):
        self.state = _INITIAL_STATE_HEATER
        self.model = MODEL_HUIZUO_WYHEAT
        super().__init__(*args, **kwargs)


@pytest.fixture(scope="function")
def huizuo(request):
    request.cls.device = DummyHuizuo()


@pytest.fixture(scope="function")
def huizuo_fan(request):
    request.cls.device = DummyHuizuoFan()


@pytest.fixture(scope="function")
def huizuo_fan2(request):
    request.cls.device = DummyHuizuoFan2()


@pytest.fixture(scope="function")
def huizuo_heater(request):
    request.cls.device = DummyHuizuoHeater()


@pytest.mark.usefixtures("huizuo")
class TestHuizuo(TestCase):
    def test_on(self):
        self.device.off()  # ensure off
        assert self.device.status().is_on is False

        self.device.on()
        assert self.device.status().is_on is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.device.status().is_on is True

        self.device.off()
        assert self.device.status().is_on is False

    def test_status(self):
        status = self.device.status()
        assert status.is_on is _INITIAL_STATE["power"]
        assert status.brightness is _INITIAL_STATE["brightness"]
        assert status.color_temp is _INITIAL_STATE["color_temp"]

    def test_brightness(self):
        def lamp_brightness():
            return self.device.status().brightness

        self.device.set_brightness(1)
        assert lamp_brightness() == 1
        self.device.set_brightness(64)
        assert lamp_brightness() == 64
        self.device.set_brightness(100)
        assert lamp_brightness() == 100

        with pytest.raises(HuizuoException):
            self.device.set_brightness(-1)

        with pytest.raises(HuizuoException):
            self.device.set_brightness(101)

    def test_color_temp(self):
        def lamp_color_temp():
            return self.device.status().color_temp

        self.device.set_color_temp(3000)
        assert lamp_color_temp() == 3000
        self.device.set_color_temp(4200)
        assert lamp_color_temp() == 4200
        self.device.set_color_temp(6400)
        assert lamp_color_temp() == 6400

        with pytest.raises(HuizuoException):
            self.device.set_color_temp(2999)

        with pytest.raises(HuizuoException):
            self.device.set_color_temp(6401)


@pytest.mark.usefixtures("huizuo_fan")
class TestHuizuoFan(TestCase):
    def test_fan_on(self):
        self.device.fan_off()  # ensure off
        assert self.device.status().is_fan_on is False

        self.device.fan_on()
        assert self.device.status().is_fan_on is True

    def test_fan_off(self):
        self.device.fan_on()  # ensure on
        assert self.device.status().is_fan_on is True

        self.device.fan_off()
        assert self.device.status().is_fan_on is False

    def test_fan_status(self):
        status = self.device.status()
        assert status.is_fan_on is _INITIAL_STATE_FAN["fan_power"]
        assert status.fan_speed_level is _INITIAL_STATE_FAN["fan_level"]
        assert status.is_fan_reverse is _INITIAL_STATE_FAN["fan_motor_reverse"]
        assert status.fan_mode is _INITIAL_STATE_FAN["fan_mode"]

    def test_fan_level(self):
        def fan_level():
            return self.device.status().fan_speed_level

        self.device.set_fan_level(0)
        assert fan_level() == 0
        self.device.set_fan_level(100)
        assert fan_level() == 100

        with pytest.raises(HuizuoException):
            self.device.set_fan_level(-1)

        with pytest.raises(HuizuoException):
            self.device.set_fan_level(101)

    def test_fan_motor_reverse(self):
        def fan_reverse():
            return self.device.status().is_fan_reverse

        self.device.fan_reverse_on()
        assert fan_reverse() is True
        self.device.fan_reverse_off()
        assert fan_reverse() is False

    def test_fan_mode(self):
        def fan_mode():
            return self.device.status().fan_mode

        self.device.set_basic_fan_mode()
        assert fan_mode() == 0
        self.device.set_natural_fan_mode()
        assert fan_mode() == 1


@pytest.mark.usefixtures("huizuo_fan2")
class TestHuizuoFan2(TestCase):
    # This device has no 'reverse' mode, so let's check this
    def test_fan_motor_reverse(self):
        with pytest.raises(HuizuoException):
            self.device.fan_reverse_on()

        with pytest.raises(HuizuoException):
            self.device.fan_reverse_off()


@pytest.mark.usefixtures("huizuo_heater")
class TestHuizuoHeater(TestCase):
    def test_heater_on(self):
        self.device.heater_off()  # ensure off
        assert self.device.status().is_heater_on is False

        self.device.heater_on()
        assert self.device.status().is_heater_on is True

    def test_heater_off(self):
        self.device.heater_on()  # ensure on
        assert self.device.status().is_heater_on is True

        self.device.heater_off()
        assert self.device.status().is_heater_on is False

    def test_heat_level(self):
        def heat_level():
            return self.device.status().heat_level

        self.device.set_heat_level(1)
        assert heat_level() == 1
        self.device.set_heat_level(3)
        assert heat_level() == 3

        with pytest.raises(HuizuoException):
            self.device.set_heat_level(0)
        with pytest.raises(HuizuoException):
            self.device.set_heat_level(4)



from unittest import TestCase

import pytest

from miio import PhilipsRwread
from miio.philips_rwread import (
    MODEL_PHILIPS_LIGHT_RWREAD,
    MotionDetectionSensitivity,
    PhilipsRwreadException,
    PhilipsRwreadStatus,
)

from .dummies import DummyDevice


class DummyPhilipsRwread(DummyDevice, PhilipsRwread):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_PHILIPS_LIGHT_RWREAD
        self.state = {
            "power": "on",
            "bright": 53,
            "dv": 0,
            "snm": 1,
            "flm": 0,
            "flmv": 2,
            "chl": 0,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_bright": lambda x: self._set_state("bright", x),
            "apply_fixed_scene": lambda x: self._set_state("snm", x),
            "delay_off": lambda x: self._set_state("dv", x),
            "enable_flm": lambda x: self._set_state("flm", x),
            "set_flmvalue": lambda x: self._set_state("flmv", x),
            "enable_chl": lambda x: self._set_state("chl", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def philips_eyecare(request):
    request.cls.device = DummyPhilipsRwread()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("philips_eyecare")
class TestPhilipsRwread(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(PhilipsRwreadStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().brightness == self.device.start_state["bright"]
        assert self.state().delay_off_countdown == self.device.start_state["dv"]
        assert self.state().scene == self.device.start_state["snm"]
        assert self.state().motion_detection is (self.device.start_state["flm"] == 1)
        assert self.state().motion_detection_sensitivity == MotionDetectionSensitivity(
            self.device.start_state["flmv"]
        )

        assert self.state().child_lock is (self.device.start_state["chl"] == 1)

    def test_set_brightness(self):
        def brightness():
            return self.device.status().brightness

        self.device.set_brightness(1)
        assert brightness() == 1
        self.device.set_brightness(50)
        assert brightness() == 50
        self.device.set_brightness(100)

        with pytest.raises(PhilipsRwreadException):
            self.device.set_brightness(-1)

        with pytest.raises(PhilipsRwreadException):
            self.device.set_brightness(0)

        with pytest.raises(PhilipsRwreadException):
            self.device.set_brightness(101)

    def test_set_scene(self):
        def scene():
            return self.device.status().scene

        self.device.set_scene(1)
        assert scene() == 1
        self.device.set_scene(2)
        assert scene() == 2

        with pytest.raises(PhilipsRwreadException):
            self.device.set_scene(-1)

        with pytest.raises(PhilipsRwreadException):
            self.device.set_scene(0)

        with pytest.raises(PhilipsRwreadException):
            self.device.set_scene(5)

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(1)
        assert delay_off_countdown() == 1
        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200

        with pytest.raises(PhilipsRwreadException):
            self.device.delay_off(-1)

    def test_set_motion_detection(self):
        def motion_detection():
            return self.device.status().motion_detection

        self.device.set_motion_detection(True)
        assert motion_detection() is True

        self.device.set_motion_detection(False)
        assert motion_detection() is False

    def test_set_motion_detection_sensitivity(self):
        def motion_detection_sensitivity():
            return self.device.status().motion_detection_sensitivity

        self.device.set_motion_detection_sensitivity(MotionDetectionSensitivity.Low)
        assert motion_detection_sensitivity() == MotionDetectionSensitivity.Low

        self.device.set_motion_detection_sensitivity(MotionDetectionSensitivity.Medium)
        assert motion_detection_sensitivity() == MotionDetectionSensitivity.Medium

        self.device.set_motion_detection_sensitivity(MotionDetectionSensitivity.High)
        assert motion_detection_sensitivity() == MotionDetectionSensitivity.High

    def test_set_child_lock(self):
        def child_lock():
            return self.device.status().child_lock

        self.device.set_child_lock(True)
        assert child_lock() is True

        self.device.set_child_lock(False)
        assert child_lock() is False



from unittest import TestCase

import pytest

from miio import PowerStrip
from miio.powerstrip import (
    MODEL_POWER_STRIP_V1,
    MODEL_POWER_STRIP_V2,
    PowerMode,
    PowerStripException,
    PowerStripStatus,
)

from .dummies import DummyDevice


class DummyPowerStripV1(DummyDevice, PowerStrip):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_POWER_STRIP_V1
        self.state = {
            "power": "on",
            "mode": "normal",
            "temperature": 32.5,
            "current": 25.5,
            "power_consume_rate": 12.5,
            "voltage": 23057,
            "power_factor": 12,
            "elec_leakage": 8,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_power_mode": lambda x: self._set_state("mode", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def powerstripv1(request):
    request.cls.device = DummyPowerStripV1()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("powerstripv1")
class TestPowerStripV1(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(PowerStripStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().mode == PowerMode(self.device.start_state["mode"])
        assert self.state().temperature == self.device.start_state["temperature"]
        assert self.state().current == self.device.start_state["current"]
        assert self.state().load_power == self.device.start_state["power_consume_rate"]
        assert self.state().voltage == self.device.start_state["voltage"] / 100.0
        assert self.state().power_factor == self.device.start_state["power_factor"]
        assert self.state().leakage_current == self.device.start_state["elec_leakage"]

    def test_status_without_power_consume_rate(self):
        self.device._reset_state()

        self.device.state["power_consume_rate"] = None
        assert self.state().load_power is None

    def test_status_without_current(self):
        self.device._reset_state()

        self.device.state["current"] = None
        assert self.state().current is None

    def test_status_without_mode(self):
        self.device._reset_state()

        # The Power Strip  2 doesn't support power modes
        self.device.state["mode"] = None
        assert self.state().mode is None

    def test_set_power_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_power_mode(PowerMode.Eco)
        assert mode() == PowerMode.Eco
        self.device.set_power_mode(PowerMode.Normal)
        assert mode() == PowerMode.Normal


class DummyPowerStripV2(DummyDevice, PowerStrip):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_POWER_STRIP_V2
        self.state = {
            "power": "on",
            "mode": "normal",
            "temperature": 32.5,
            "current": 25.5,
            "power_consume_rate": 12.5,
            "wifi_led": "off",
            "power_price": 49,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_power_mode": lambda x: self._set_state("mode", x),
            "set_wifi_led": lambda x: self._set_state("wifi_led", x),
            "set_power_price": lambda x: self._set_state("power_price", x),
            "set_rt_power": lambda x: True,
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def powerstripv2(request):
    request.cls.device = DummyPowerStripV2()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("powerstripv2")
class TestPowerStripV2(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(PowerStripStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().mode == PowerMode(self.device.start_state["mode"])
        assert self.state().temperature == self.device.start_state["temperature"]
        assert self.state().current == self.device.start_state["current"]
        assert self.state().load_power == self.device.start_state["power_consume_rate"]
        assert self.state().voltage is None
        assert self.state().power_factor is None
        assert self.state().leakage_current is None

    def test_status_without_power_consume_rate(self):
        self.device._reset_state()

        self.device.state["power_consume_rate"] = None
        assert self.state().load_power is None

    def test_status_without_current(self):
        self.device._reset_state()

        self.device.state["current"] = None
        assert self.state().current is None

    def test_status_without_mode(self):
        self.device._reset_state()

        # The Power Strip  2 doesn't support power modes
        self.device.state["mode"] = None
        assert self.state().mode is None

    def test_set_power_mode(self):
        def mode():
            return self.device.status().mode

        self.device.set_power_mode(PowerMode.Eco)
        assert mode() == PowerMode.Eco
        self.device.set_power_mode(PowerMode.Normal)
        assert mode() == PowerMode.Normal

    def test_set_wifi_led(self):
        def wifi_led():
            return self.device.status().wifi_led

        self.device.set_wifi_led(True)
        assert wifi_led() is True

        self.device.set_wifi_led(False)
        assert wifi_led() is False

    def test_set_power_price(self):
        def power_price():
            return self.device.status().power_price

        self.device.set_power_price(0)
        assert power_price() == 0
        self.device.set_power_price(1)
        assert power_price() == 1
        self.device.set_power_price(2)
        assert power_price() == 2

        with pytest.raises(PowerStripException):
            self.device.set_power_price(-1)

        with pytest.raises(PowerStripException):
            self.device.set_power_price(1000)

    def test_status_without_power_price(self):
        self.device._reset_state()

        self.device.state["power_price"] = None
        assert self.state().power_price is None

    def test_set_realtime_power(self):
        """The method is open-loop.

        The new state cannot be retrieved.
        """
        self.device.set_realtime_power(True)
        self.device.set_realtime_power(False)



import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .click_common import command, format_output
from .device import Device, DeviceStatus

_LOGGER = logging.getLogger(__name__)

MODEL_PWZN_RELAY_APPLE = "pwzn.relay.apple"
MODEL_PWZN_RELAY_BANANA = "pwzn.relay.banana"

AVAILABLE_PROPERTIES = {
    MODEL_PWZN_RELAY_APPLE: [
        "relay_status",
        "on_count",
        "name0",
        "name1",
        "name2",
        "name3",
        "name4",
        "name5",
        "name6",
        "name7",
        "name8",
        "name9",
        "name10",
        "name11",
        "name12",
        "name13",
        "name14",
        "name15",
    ],
    MODEL_PWZN_RELAY_BANANA: [
        "relay_status",
        "on_count",
        "name0",
        "name1",
        "name2",
        "name3",
        "name4",
        "name5",
        "name6",
        "name7",
        "name8",
        "name9",
        "name10",
        "name11",
        "name12",
        "name13",
        "name14",
        "name15",
    ],
}


class PwznRelayStatus(DeviceStatus):
    """Container for status reports from the plug."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Response of a PWZN Relay Apple (pwzn.relay.apple)

        { 'relay_status': 9, 'on_count': 2, 'name0': 'channel1', 'name1': '',
        'name2': '', 'name3': '', 'name4': '', 'name5': '', 'name6': '',
        'name7': '', 'name8': '', 'name9': '', 'name10': '', 'name11': '',
        'name12': '', 'name13': '', 'name14': '', 'name15': '' }
        """
        self.data = data

    @property
    def relay_state(self) -> Optional[int]:
        """Current relay state."""
        if "relay_status" in self.data:
            return self.data["relay_status"]
        return None

    @property
    def relay_names(self) -> Dict[int, str]:
        def _extract_index_from_key(name) -> int:
            """extract the index from the variable."""
            return int(name[4:])

        return {
            _extract_index_from_key(name): value
            for name, value in self.data.items()
            if name.startswith("name")
        }

    @property
    def on_count(self) -> Optional[int]:
        """Number of on relay."""
        if "on_count" in self.data:
            return self.data["on_count"]
        return None


class PwznRelay(Device):
    """Main class representing the PWZN Relay."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_PWZN_RELAY_APPLE,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_PWZN_RELAY_APPLE

    @command(default_output=format_output("", "on_count: {result.on_count}\n"))
    def status(self) -> PwznRelayStatus:
        """Retrieve properties."""

        properties = AVAILABLE_PROPERTIES[self.model].copy()
        values = self.get_properties(properties)

        return PwznRelayStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(
        click.argument("number", type=int),
        default_output=format_output("Turn on relay {number}"),
    )
    def relay_on(self, number: int = 0):
        """Relay X on."""
        if self.send("power_on", [number]) == [0]:
            return ["ok"]

    @command(
        click.argument("number", type=int),
        default_output=format_output("Turn off relay {number}"),
    )
    def relay_off(self, number: int = 0):
        """Relay X off."""
        if self.send("power_off", [number]) == [0]:
            return ["ok"]

    @command(default_output=format_output("Turn on all relay"))
    def all_relay_on(self):
        """Relay all on."""
        return self.send("power_all", [1])

    @command(default_output=format_output("Turn off all relay"))
    def all_relay_off(self):
        """Relay all off."""
        return self.send("power_all", [0])

    @command(
        click.argument("number", type=int),
        click.argument("name", type=str),
        default_output=format_output("Set relay {number} name to {name}"),
    )
    def set_name(self, number: int = 0, name: str = ""):
        """Set relay X name."""
        return self.send("set_name", [number, name])



from unittest import TestCase

import pytest

from miio import PhilipsBulb, PhilipsWhiteBulb
from miio.philips_bulb import (
    MODEL_PHILIPS_LIGHT_BULB,
    MODEL_PHILIPS_LIGHT_HBULB,
    PhilipsBulbException,
    PhilipsBulbStatus,
)

from .dummies import DummyDevice


class DummyPhilipsBulb(DummyDevice, PhilipsBulb):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_PHILIPS_LIGHT_BULB
        self.state = {"power": "on", "bright": 100, "cct": 10, "snm": 0, "dv": 0}
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_bright": lambda x: self._set_state("bright", x),
            "set_cct": lambda x: self._set_state("cct", x),
            "delay_off": lambda x: self._set_state("dv", x),
            "apply_fixed_scene": lambda x: self._set_state("snm", x),
            "set_bricct": lambda x: (
                self._set_state("bright", [x[0]]),
                self._set_state("cct", [x[1]]),
            ),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def philips_bulb(request):
    request.cls.device = DummyPhilipsBulb()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("philips_bulb")
class TestPhilipsBulb(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(PhilipsBulbStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().brightness == self.device.start_state["bright"]
        assert self.state().color_temperature == self.device.start_state["cct"]
        assert self.state().scene == self.device.start_state["snm"]
        assert self.state().delay_off_countdown == self.device.start_state["dv"]

    def test_set_brightness(self):
        def brightness():
            return self.device.status().brightness

        self.device.set_brightness(1)
        assert brightness() == 1
        self.device.set_brightness(50)
        assert brightness() == 50
        self.device.set_brightness(100)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness(-1)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness(0)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness(101)

    def test_set_color_temperature(self):
        def color_temperature():
            return self.device.status().color_temperature

        self.device.set_color_temperature(20)
        assert color_temperature() == 20
        self.device.set_color_temperature(30)
        assert color_temperature() == 30
        self.device.set_color_temperature(10)

        with pytest.raises(PhilipsBulbException):
            self.device.set_color_temperature(-1)

        with pytest.raises(PhilipsBulbException):
            self.device.set_color_temperature(0)

        with pytest.raises(PhilipsBulbException):
            self.device.set_color_temperature(101)

    def test_set_brightness_and_color_temperature(self):
        def color_temperature():
            return self.device.status().color_temperature

        def brightness():
            return self.device.status().brightness

        self.device.set_brightness_and_color_temperature(20, 21)
        assert brightness() == 20
        assert color_temperature() == 21
        self.device.set_brightness_and_color_temperature(31, 30)
        assert brightness() == 31
        assert color_temperature() == 30
        self.device.set_brightness_and_color_temperature(10, 11)
        assert brightness() == 10
        assert color_temperature() == 11

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness_and_color_temperature(-1, 10)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness_and_color_temperature(10, -1)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness_and_color_temperature(0, 10)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness_and_color_temperature(10, 0)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness_and_color_temperature(101, 10)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness_and_color_temperature(10, 101)

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200

        with pytest.raises(PhilipsBulbException):
            self.device.delay_off(-1)

        with pytest.raises(PhilipsBulbException):
            self.device.delay_off(0)

    def test_set_scene(self):
        def scene():
            return self.device.status().scene

        self.device.set_scene(1)
        assert scene() == 1
        self.device.set_scene(2)
        assert scene() == 2

        with pytest.raises(PhilipsBulbException):
            self.device.set_scene(-1)

        with pytest.raises(PhilipsBulbException):
            self.device.set_scene(0)

        with pytest.raises(PhilipsBulbException):
            self.device.set_scene(5)


class DummyPhilipsWhiteBulb(DummyDevice, PhilipsWhiteBulb):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_PHILIPS_LIGHT_HBULB
        self.state = {"power": "on", "bri": 100, "dv": 0}
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_bright": lambda x: self._set_state("bri", x),
            "delay_off": lambda x: self._set_state("dv", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def philips_white_bulb(request):
    request.cls.device = DummyPhilipsWhiteBulb()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("philips_white_bulb")
class TestPhilipsWhiteBulb(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(PhilipsBulbStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().brightness == self.device.start_state["bri"]
        assert self.state().delay_off_countdown == self.device.start_state["dv"]
        assert self.state().color_temperature is None
        assert self.state().scene is None

    def test_set_brightness(self):
        def brightness():
            return self.device.status().brightness

        self.device.set_brightness(1)
        assert brightness() == 1
        self.device.set_brightness(50)
        assert brightness() == 50
        self.device.set_brightness(100)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness(-1)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness(0)

        with pytest.raises(PhilipsBulbException):
            self.device.set_brightness(101)

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200

        with pytest.raises(PhilipsBulbException):
            self.device.delay_off(-1)

        with pytest.raises(PhilipsBulbException):
            self.device.delay_off(0)



import json
import logging
import sqlite3
import tempfile
from pprint import pformat as pf
from typing import Iterator

import attr
import click
import defusedxml.ElementTree as ET
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

logging.basicConfig(level=logging.INFO)
_LOGGER = logging.getLogger(__name__)


@attr.s
class DeviceConfig:
    """A presentation of a device including its name, model, ip etc."""

    name = attr.ib()
    mac = attr.ib()
    ip = attr.ib()
    token = attr.ib()
    model = attr.ib()
    everything = attr.ib(default=None)


def read_android_yeelight(db) -> Iterator[DeviceConfig]:
    """Read tokens from Yeelight's android backup."""
    _LOGGER.info("Reading tokens from Yeelight Android DB")
    xml = ET.parse(db)
    devicelist = xml.find(".//set[@name='deviceList']")
    if not devicelist:
        _LOGGER.warning("Unable to find deviceList")
        return []

    for dev_elem in list(devicelist):
        dev = json.loads(dev_elem.text)
        ip = dev["localip"]
        mac = dev["mac"]
        model = dev["model"]
        name = dev["name"]
        token = dev["token"]

        config = DeviceConfig(
            name=name, ip=ip, mac=mac, model=model, token=token, everything=dev
        )

        yield config


class BackupDatabaseReader:
    """Main class for reading backup files.

    Example:
    .. code-block:: python

        r = BackupDatabaseReader()
        devices = r.read_tokens("/tmp/database.sqlite")
        for dev in devices:
            print("Got %s with token %s" % (dev.ip, dev.token)
    """

    def __init__(self, dump_raw=False):
        self.dump_raw = dump_raw

    @staticmethod
    def dump_raw(dev):
        """Dump whole database."""
        raw = {k: dev[k] for k in dev.keys()}
        _LOGGER.info(pf(raw))

    @staticmethod
    def decrypt_ztoken(ztoken):
        """Decrypt the given ztoken, used by apple."""
        if ztoken is None or len(ztoken) <= 32:
            return str(ztoken)

        keystring = "00000000000000000000000000000000"
        key = bytes.fromhex(keystring)
        cipher = Cipher(  # nosec
            algorithms.AES(key), modes.ECB(), backend=default_backend()
        )
        decryptor = cipher.decryptor()
        token = decryptor.update(bytes.fromhex(ztoken[:64])) + decryptor.finalize()

        return token.decode()

    def read_apple(self) -> Iterator[DeviceConfig]:
        """Read Apple-specific database file."""
        _LOGGER.info("Reading tokens from Apple DB")
        c = self.conn.execute("SELECT * FROM ZDEVICE WHERE ZTOKEN IS NOT '';")
        for dev in c.fetchall():
            if self.dump_raw:
                BackupDatabaseReader.dump_raw(dev)
            ip = dev["ZLOCALIP"]
            mac = dev["ZMAC"]
            model = dev["ZMODEL"]
            name = dev["ZNAME"]
            token = BackupDatabaseReader.decrypt_ztoken(dev["ZTOKEN"])

            config = DeviceConfig(
                name=name, mac=mac, ip=ip, model=model, token=token, everything=dev
            )
            yield config

    def read_android(self) -> Iterator[DeviceConfig]:
        """Read Android-specific database file."""
        _LOGGER.info("Reading tokens from Android DB")
        c = self.conn.execute("SELECT * FROM devicerecord WHERE token IS NOT '';")
        for dev in c.fetchall():
            if self.dump_raw:
                BackupDatabaseReader.dump_raw(dev)
            ip = dev["localIP"]
            mac = dev["mac"]
            model = dev["model"]
            name = dev["name"]
            token = dev["token"]

            config = DeviceConfig(
                name=name, ip=ip, mac=mac, model=model, token=token, everything=dev
            )
            yield config

    def read_tokens(self, db) -> Iterator[DeviceConfig]:
        """Read device information out from a given database file.

        :param str db: Database file
        """
        self.db = db
        _LOGGER.info("Reading database from %s" % db)
        self.conn = sqlite3.connect(db)

        self.conn.row_factory = sqlite3.Row
        with self.conn:
            is_android = (
                self.conn.execute(
                    "SELECT name FROM sqlite_master WHERE type='table' AND name='devicerecord';"
                ).fetchone()
                is not None
            )
            is_apple = (
                self.conn.execute(
                    "SELECT name FROM sqlite_master WHERE type='table' AND name='ZDEVICE'"
                ).fetchone()
                is not None
            )
            if is_android:
                yield from self.read_android()
            elif is_apple:
                yield from self.read_apple()
            else:
                _LOGGER.error("Error, unknown database type!")


@click.command()
@click.argument("backup")
@click.option(
    "--write-to-disk",
    type=click.File("wb"),
    help="writes sqlite3 db to a file for debugging",
)
@click.option(
    "--password", type=str, help="password if the android database is encrypted"
)
@click.option(
    "--dump-all", is_flag=True, default=False, help="dump devices without ip addresses"
)
@click.option("--dump-raw", is_flag=True, help="dumps raw rows")
def main(backup, write_to_disk, password, dump_all, dump_raw):
    """Reads device information out from an sqlite3 DB.

    If the given file is an Android backup (.ab), the database will be extracted
    automatically. If the given file is an iOS backup, the tokens will be extracted (and
    decrypted if needed) automatically.
    """

    def read_miio_database(tar):
        DBFILE = "apps/com.xiaomi.smarthome/db/miio2.db"
        try:
            db = tar.extractfile(DBFILE)
        except KeyError as ex:
            click.echo("Unable to find miio database file %s: %s" % (DBFILE, ex))
            return []
        if write_to_disk:
            file = write_to_disk
        else:
            file = tempfile.NamedTemporaryFile()
        with file as fp:
            click.echo("Saving database to %s" % fp.name)
            fp.write(db.read())

            return list(reader.read_tokens(fp.name))

    def read_yeelight_database(tar):
        DBFILE = "apps/com.yeelight.cherry/sp/miot.xml"
        _LOGGER.info("Trying to read %s", DBFILE)
        try:
            db = tar.extractfile(DBFILE)
        except KeyError as ex:
            click.echo("Unable to find yeelight database file %s: %s" % (DBFILE, ex))
            return []

        return list(read_android_yeelight(db))

    devices = []
    reader = BackupDatabaseReader(dump_raw)
    if backup.endswith(".ab"):
        try:
            from android_backup import AndroidBackup
        except ModuleNotFoundError:
            click.echo(
                "You need to install android_backup to extract "
                "tokens from Android backup files."
            )
            return

        with AndroidBackup(backup, stream=False) as f:
            tar = f.read_data(password)

            devices.extend(read_miio_database(tar))

            devices.extend(read_yeelight_database(tar))
    else:
        devices = list(reader.read_tokens(backup))

    for dev in devices:
        if dev.ip or dump_all:
            click.echo(
                "%s\n"
                "\tModel: %s\n"
                "\tIP address: %s\n"
                "\tToken: %s\n"
                "\tMAC: %s" % (dev.name, dev.model, dev.ip, dev.token, dev.mac)
            )
            if dump_raw:
                click.echo(dev)


if __name__ == "__main__":
    main()



import enum
import logging
from typing import Any, Dict

import click

from .airhumidifier import AirHumidifierException
from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus

_LOGGER = logging.getLogger(__name__)

# Xiaomi Zero Fog Humidifier
MODEL_HUMIDIFIER_JSQ001 = "shuii.humidifier.jsq001"

# Array of properties in same order as in humidifier response
AVAILABLE_PROPERTIES = {
    MODEL_HUMIDIFIER_JSQ001: [
        "temperature",  # (degrees, int)
        "humidity",  # (percentage, int)
        "mode",  # ( 0: Intelligent, 1: Level1, ..., 5:Level4)
        "buzzer",  # (0: off, 1: on)
        "child_lock",  # (0: off, 1: on)
        "led_brightness",  # (0: off, 1: low, 2: high)
        "power",  # (0: off, 1: on)
        "no_water",  # (0: enough, 1: add water)
        "lid_opened",  # (0: ok, 1: lid is opened)
    ]
}


class OperationMode(enum.Enum):
    Intelligent = 0
    Level1 = 1
    Level2 = 2
    Level3 = 3
    Level4 = 4


class LedBrightness(enum.Enum):
    Off = 0
    Low = 1
    High = 2


class AirHumidifierStatus(DeviceStatus):
    """Container for status reports from the air humidifier jsq."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Status of an Air Humidifier (shuii.humidifier.jsq001):

            [24, 30, 1, 1, 0, 2, 0, 0, 0]

        Parsed by AirHumidifierJsq device as:
            {'temperature': 24, 'humidity': 29, 'mode': 1, 'buzzer': 1,
             'child_lock': 0, 'led_brightness': 2, 'power': 0, 'no_water': 0,
              'lid_opened': 0}
        """
        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return "on" if self.data["power"] == 1 else "off"

    @property
    def is_on(self) -> bool:
        """True if device is turned on."""
        return self.power == "on"

    @property
    def mode(self) -> OperationMode:
        """Operation mode.

        Can be either low, medium, high or humidity.
        """

        try:
            mode = OperationMode(self.data["mode"])
        except ValueError as e:
            _LOGGER.exception("Cannot parse mode: %s", e)
            return OperationMode.Intelligent

        return mode

    @property
    def temperature(self) -> int:
        """Current temperature in degree celsius."""
        return self.data["temperature"]

    @property
    def humidity(self) -> int:
        """Current humidity in percent."""
        return self.data["humidity"]

    @property
    def buzzer(self) -> bool:
        """True if buzzer is turned on."""
        return self.data["buzzer"] == 1

    @property
    def led_brightness(self) -> LedBrightness:
        """Buttons illumination Brightness level."""
        try:
            brightness = LedBrightness(self.data["led_brightness"])
        except ValueError as e:
            _LOGGER.exception("Cannot parse brightness: %s", e)
            return LedBrightness.Off

        return brightness

    @property
    def led(self) -> bool:
        """True if LED is turned on."""
        return self.led_brightness is not LedBrightness.Off

    @property
    def child_lock(self) -> bool:
        """Return True if child lock is on."""
        return self.data["child_lock"] == 1

    @property
    def no_water(self) -> bool:
        """True if the water tank is empty."""
        return self.data["no_water"] == 1

    @property
    def lid_opened(self) -> bool:
        """True if the water tank is detached."""
        return self.data["lid_opened"] == 1


class AirHumidifierJsq(Device):
    """Implementation of Xiaomi Zero Fog Humidifier: shuii.humidifier.jsq001."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_HUMIDIFIER_JSQ001,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        self.model = model if model in AVAILABLE_PROPERTIES else MODEL_HUMIDIFIER_JSQ001

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode}\n"
            "Temperature: {result.temperature} °C\n"
            "Humidity: {result.humidity} %\n"
            "Buzzer: {result.buzzer}\n"
            "LED brightness: {result.led_brightness}\n"
            "Child lock: {result.child_lock}\n"
            "No water: {result.no_water}\n"
            "Lid opened: {result.lid_opened}\n",
        )
    )
    def status(self) -> AirHumidifierStatus:
        """Retrieve properties."""

        values = self.send("get_props")

        # Response of an Air Humidifier (shuii.humidifier.jsq001):
        # [24, 37, 3, 1, 0, 2, 0, 0, 0]
        #
        # status[0] : temperature (degrees, int)
        # status[1]: humidity (percentage, int)
        # status[2]: mode ( 0: Intelligent, 1: Level1, ..., 5:Level4)
        # status[3]: buzzer (0: off, 1: on)
        # status[4]: lock (0: off, 1: on)
        # status[5]: brightness (0: off, 1: low, 2: high)
        # status[6]: power (0: off, 1: on)
        # status[7]: water level state (0: ok, 1: add water)
        # status[8]: lid state (0: ok, 1: lid is opened)

        properties = AVAILABLE_PROPERTIES[self.model]
        if len(properties) != len(values):
            _LOGGER.error(
                "Count (%s) of requested properties (%s) does not match the "
                "count (%s) of received values (%s).",
                len(properties),
                properties,
                len(values),
                values,
            )

        return AirHumidifierStatus({k: v for k, v in zip(properties, values)})

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_start", [1])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_start", [0])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        value = mode.value
        if value not in (om.value for om in OperationMode):
            raise AirHumidifierException(
                "{} is not a valid OperationMode value".format(value)
            )

        return self.send("set_mode", [value])

    @command(
        click.argument("brightness", type=EnumType(LedBrightness)),
        default_output=format_output("Setting LED brightness to {brightness}"),
    )
    def set_led_brightness(self, brightness: LedBrightness):
        """Set led brightness."""
        value = brightness.value
        if value not in (lb.value for lb in LedBrightness):
            raise AirHumidifierException(
                "{} is not a valid LedBrightness value".format(value)
            )

        return self.send("set_brightness", [value])

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        brightness = LedBrightness.High if led else LedBrightness.Off
        return self.set_led_brightness(brightness)

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        return self.send("set_buzzer", [int(bool(buzzer))])

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        return self.send("set_lock", [int(bool(lock))])



from unittest import TestCase

import pytest

from miio import PhilipsEyecare
from miio.philips_eyecare import PhilipsEyecareException, PhilipsEyecareStatus

from .dummies import DummyDevice


class DummyPhilipsEyecare(DummyDevice, PhilipsEyecare):
    def __init__(self, *args, **kwargs):
        self.state = {
            "power": "on",
            "bright": 100,
            "notifystatus": "off",
            "ambstatus": "off",
            "ambvalue": 100,
            "eyecare": "on",
            "scene_num": 3,
            "bls": "on",
            "dvalue": 0,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_eyecare": lambda x: self._set_state("eyecare", x),
            "set_bright": lambda x: self._set_state("bright", x),
            "set_user_scene": lambda x: self._set_state("scene_num", x),
            "delay_off": lambda x: self._set_state("dvalue", x),
            "enable_bl": lambda x: self._set_state("bls", x),
            "set_notifyuser": lambda x: self._set_state("notifystatus", x),
            "enable_amb": lambda x: self._set_state("ambstatus", x),
            "set_amb_bright": lambda x: self._set_state("ambvalue", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def philips_eyecare(request):
    request.cls.device = DummyPhilipsEyecare()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("philips_eyecare")
class TestPhilipsEyecare(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(PhilipsEyecareStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().brightness == self.device.start_state["bright"]
        assert self.state().reminder is (
            self.device.start_state["notifystatus"] == "on"
        )
        assert self.state().ambient is (self.device.start_state["ambstatus"] == "on")
        assert self.state().ambient_brightness == self.device.start_state["ambvalue"]
        assert self.state().eyecare is (self.device.start_state["eyecare"] == "on")
        assert self.state().scene == self.device.start_state["scene_num"]
        assert self.state().smart_night_light is (
            self.device.start_state["bls"] == "on"
        )
        assert self.state().delay_off_countdown == self.device.start_state["dvalue"]

    def test_eyecare(self):
        def eyecare():
            return self.device.status().eyecare

        self.device.eyecare_on()
        assert eyecare() is True
        self.device.eyecare_off()
        assert eyecare() is False

    def test_set_brightness(self):
        def brightness():
            return self.device.status().brightness

        self.device.set_brightness(1)
        assert brightness() == 1
        self.device.set_brightness(50)
        assert brightness() == 50
        self.device.set_brightness(100)

        with pytest.raises(PhilipsEyecareException):
            self.device.set_brightness(-1)

        with pytest.raises(PhilipsEyecareException):
            self.device.set_brightness(0)

        with pytest.raises(PhilipsEyecareException):
            self.device.set_brightness(101)

    def test_set_scene(self):
        def scene():
            return self.device.status().scene

        self.device.set_scene(1)
        assert scene() == 1
        self.device.set_scene(2)
        assert scene() == 2

        with pytest.raises(PhilipsEyecareException):
            self.device.set_scene(-1)

        with pytest.raises(PhilipsEyecareException):
            self.device.set_scene(0)

        with pytest.raises(PhilipsEyecareException):
            self.device.set_scene(5)

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(1)
        assert delay_off_countdown() == 1
        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200

        with pytest.raises(PhilipsEyecareException):
            self.device.delay_off(-1)

    def test_smart_night_light(self):
        def smart_night_light():
            return self.device.status().smart_night_light

        self.device.smart_night_light_on()
        assert smart_night_light() is True
        self.device.smart_night_light_off()
        assert smart_night_light() is False

    def test_reminder(self):
        def reminder():
            return self.device.status().reminder

        self.device.reminder_on()
        assert reminder() is True
        self.device.reminder_off()
        assert reminder() is False

    def test_ambient(self):
        def ambient():
            return self.device.status().ambient

        self.device.ambient_on()
        assert ambient() is True
        self.device.ambient_off()
        assert ambient() is False

    def test_set_ambient_brightness(self):
        def ambient_brightness():
            return self.device.status().ambient_brightness

        self.device.set_ambient_brightness(1)
        assert ambient_brightness() == 1
        self.device.set_ambient_brightness(50)
        assert ambient_brightness() == 50
        self.device.set_ambient_brightness(100)

        with pytest.raises(PhilipsEyecareException):
            self.device.set_ambient_brightness(-1)

        with pytest.raises(PhilipsEyecareException):
            self.device.set_ambient_brightness(0)

        with pytest.raises(PhilipsEyecareException):
            self.device.set_ambient_brightness(101)



import enum
import logging
import random
from typing import Any, Optional

from .click_common import command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_ACPARTNER_MCN02 = "lumi.acpartner.mcn02"


class AirConditioningCompanionException(DeviceException):
    pass


class OperationMode(enum.Enum):
    Cool = "cool"
    Heat = "heat"
    Auto = "auto"
    Ventilate = "wind"
    Dehumidify = "dry"


class FanSpeed(enum.Enum):
    Auto = "auto_fan"
    Low = "small_fan"
    Medium = "medium_fan"
    High = "large_fan"


class SwingMode(enum.Enum):
    On = "on"
    Off = "off"


class AirConditioningCompanionStatus(DeviceStatus):
    """Container for status reports of the Xiaomi AC Companion."""

    def __init__(self, data):
        """Status constructor.

        Example response (lumi.acpartner.mcn02):
        * ['power', 'mode', 'tar_temp', 'fan_level', 'ver_swing', 'load_power']
        * ['on',    'dry',   16,        'small_fan', 'off',        84.0]
        """
        self.data = data

    @property
    def load_power(self) -> int:
        """Current power load of the air conditioner."""
        return int(self.data[-1])

    @property
    def power(self) -> str:
        """Current power state."""
        return self.data[0]

    @property
    def is_on(self) -> bool:
        """True if the device is turned on."""
        return self.power == "on"

    @property
    def mode(self) -> Optional[OperationMode]:
        """Current operation mode."""
        try:
            mode = self.data[1]
            return OperationMode(mode)
        except TypeError:
            return None

    @property
    def target_temperature(self) -> Optional[int]:
        """Target temperature."""
        try:
            return self.data[2]
        except TypeError:
            return None

    @property
    def fan_speed(self) -> Optional[FanSpeed]:
        """Current fan speed."""
        try:
            speed = self.data[3]
            return FanSpeed(speed)
        except TypeError:
            return None

    @property
    def swing_mode(self) -> Optional[SwingMode]:
        """Current swing mode."""
        try:
            mode = self.data[4]
            return SwingMode(mode)
        except TypeError:
            return None


class AirConditioningCompanionMcn02(Device):
    """Main class representing Xiaomi Air Conditioning Companion V1 and V2."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = None,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_ACPARTNER_MCN02,
    ) -> None:
        if start_id is None:
            start_id = random.randint(0, 999)  # nosec
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model != MODEL_ACPARTNER_MCN02:
            _LOGGER.error(
                "Device model %s unsupported. Please use AirConditioningCompanion",
                model,
            )

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Load power: {result.load_power}\n"
            "Target temperature: {result.target_temperature} °C\n"
            "Swing mode: {result.swing_mode}\n"
            "Fan speed: {result.fan_speed}\n"
            "Mode: {result.mode}\n",
        )
    )
    def status(self) -> AirConditioningCompanionStatus:
        """Return device status."""
        data = self.send(
            "get_prop",
            ["power", "mode", "tar_temp", "fan_level", "ver_swing", "load_power"],
        )
        return AirConditioningCompanionStatus(data)

    @command(default_output=format_output("Powering the air condition on"))
    def on(self):
        """Turn the air condition on by infrared."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering the air condition off"))
    def off(self):
        """Turn the air condition off by infrared."""
        return self.send("set_power", ["off"])

    @command(
        default_output=format_output("Sending a command to the air conditioner"),
    )
    def send_command(self, command: str, parameters: Any = None) -> Any:
        """Send a command to the air conditioner.

        :param str command: Command to execute
        """
        return self.send(command, parameters)



import codecs
import inspect
import logging
import time
from functools import partial
from ipaddress import ip_address
from typing import Callable, Dict, Optional, Type, Union  # noqa: F401

import zeroconf

from . import (
    AirConditionerMiot,
    AirConditioningCompanion,
    AirConditioningCompanionMcn02,
    AirDogX3,
    AirDogX5,
    AirDogX7SM,
    AirFresh,
    AirFreshT2017,
    AirHumidifier,
    AirHumidifierJsq,
    AirHumidifierMjjsq,
    AirPurifier,
    AirPurifierMiot,
    AirQualityMonitor,
    AqaraCamera,
    Ceil,
    ChuangmiCamera,
    ChuangmiIr,
    ChuangmiPlug,
    Cooker,
    Device,
    Fan,
    FanLeshow,
    FanMiot,
    Gateway,
    Heater,
    PhilipsBulb,
    PhilipsEyecare,
    PhilipsMoonlight,
    PhilipsRwread,
    PhilipsWhiteBulb,
    PowerStrip,
    Toiletlid,
    Vacuum,
    ViomiVacuum,
    WaterPurifier,
    WaterPurifierYunmi,
    WifiRepeater,
    WifiSpeaker,
    Yeelight,
)
from .airconditioningcompanion import (
    MODEL_ACPARTNER_V1,
    MODEL_ACPARTNER_V2,
    MODEL_ACPARTNER_V3,
)
from .airconditioningcompanionMCN import MODEL_ACPARTNER_MCN02
from .airfresh import MODEL_AIRFRESH_VA2, MODEL_AIRFRESH_VA4
from .airhumidifier import (
    MODEL_HUMIDIFIER_CA1,
    MODEL_HUMIDIFIER_CB1,
    MODEL_HUMIDIFIER_V1,
)
from .airhumidifier_mjjsq import MODEL_HUMIDIFIER_JSQ1, MODEL_HUMIDIFIER_MJJSQ
from .airqualitymonitor import (
    MODEL_AIRQUALITYMONITOR_B1,
    MODEL_AIRQUALITYMONITOR_S1,
    MODEL_AIRQUALITYMONITOR_V1,
)
from .alarmclock import AlarmClock
from .chuangmi_plug import (
    MODEL_CHUANGMI_PLUG_HMI205,
    MODEL_CHUANGMI_PLUG_HMI206,
    MODEL_CHUANGMI_PLUG_M1,
    MODEL_CHUANGMI_PLUG_M3,
    MODEL_CHUANGMI_PLUG_V1,
    MODEL_CHUANGMI_PLUG_V2,
    MODEL_CHUANGMI_PLUG_V3,
)
from .fan import (
    MODEL_FAN_P5,
    MODEL_FAN_SA1,
    MODEL_FAN_V2,
    MODEL_FAN_V3,
    MODEL_FAN_ZA1,
    MODEL_FAN_ZA3,
    MODEL_FAN_ZA4,
)
from .fan_miot import MODEL_FAN_1C, MODEL_FAN_P9, MODEL_FAN_P10, MODEL_FAN_P11
from .heater import MODEL_HEATER_MA1, MODEL_HEATER_ZA1
from .powerstrip import MODEL_POWER_STRIP_V1, MODEL_POWER_STRIP_V2
from .toiletlid import MODEL_TOILETLID_V1

_LOGGER = logging.getLogger(__name__)


DEVICE_MAP: Dict[str, Union[Type[Device], partial]] = {
    "rockrobo-vacuum-v1": Vacuum,
    "roborock-vacuum-s5": Vacuum,
    "roborock-vacuum-m1s": Vacuum,
    "roborock-vacuum-a10": Vacuum,
    "chuangmi-plug-m1": partial(ChuangmiPlug, model=MODEL_CHUANGMI_PLUG_M1),
    "chuangmi-plug-m3": partial(ChuangmiPlug, model=MODEL_CHUANGMI_PLUG_M3),
    "chuangmi-plug-v1": partial(ChuangmiPlug, model=MODEL_CHUANGMI_PLUG_V1),
    "chuangmi-plug-v2": partial(ChuangmiPlug, model=MODEL_CHUANGMI_PLUG_V2),
    "chuangmi-plug-v3": partial(ChuangmiPlug, model=MODEL_CHUANGMI_PLUG_V3),
    "chuangmi-plug-hmi205": partial(ChuangmiPlug, model=MODEL_CHUANGMI_PLUG_HMI205),
    "chuangmi-plug-hmi206": partial(ChuangmiPlug, model=MODEL_CHUANGMI_PLUG_HMI206),
    "chuangmi-plug_": partial(ChuangmiPlug, model=MODEL_CHUANGMI_PLUG_V1),
    "qmi-powerstrip-v1": partial(PowerStrip, model=MODEL_POWER_STRIP_V1),
    "zimi-powerstrip-v2": partial(PowerStrip, model=MODEL_POWER_STRIP_V2),
    "zimi-clock-myk01": AlarmClock,
    "xiaomi.aircondition.mc1": AirConditionerMiot,
    "xiaomi.aircondition.mc2": AirConditionerMiot,
    "xiaomi.aircondition.mc4": AirConditionerMiot,
    "xiaomi.aircondition.mc5": AirConditionerMiot,
    "airdog-airpurifier-x3": AirDogX3,
    "airdog-airpurifier-x5": AirDogX5,
    "airdog-airpurifier-x7sm": AirDogX7SM,
    "zhimi-airpurifier-m1": AirPurifier,  # mini model
    "zhimi-airpurifier-m2": AirPurifier,  # mini model 2
    "zhimi-airpurifier-ma1": AirPurifier,  # ms model
    "zhimi-airpurifier-ma2": AirPurifier,  # ms model 2
    "zhimi-airpurifier-sa1": AirPurifier,  # super model
    "zhimi-airpurifier-sa2": AirPurifier,  # super model 2
    "zhimi-airpurifier-v1": AirPurifier,  # v1
    "zhimi-airpurifier-v2": AirPurifier,  # v2
    "zhimi-airpurifier-v3": AirPurifier,  # v3
    "zhimi-airpurifier-v5": AirPurifier,  # v5
    "zhimi-airpurifier-v6": AirPurifier,  # v6
    "zhimi-airpurifier-v7": AirPurifier,  # v7
    "zhimi-airpurifier-mc1": AirPurifier,  # mc1
    "zhimi-airpurifier-mb3": AirPurifierMiot,  # mb3 (3/3H)
    "zhimi-airpurifier-ma4": AirPurifierMiot,  # ma4 (3)
    "chuangmi-camera-ipc009": ChuangmiCamera,
    "chuangmi-camera-ipc019": ChuangmiCamera,
    "chuangmi-ir-v2": ChuangmiIr,
    "chuangmi-remote-h102a03_": ChuangmiIr,
    "zhimi-humidifier-v1": partial(AirHumidifier, model=MODEL_HUMIDIFIER_V1),
    "zhimi-humidifier-ca1": partial(AirHumidifier, model=MODEL_HUMIDIFIER_CA1),
    "zhimi-humidifier-cb1": partial(AirHumidifier, model=MODEL_HUMIDIFIER_CB1),
    "shuii-humidifier-jsq001": partial(AirHumidifierJsq, model=MODEL_HUMIDIFIER_MJJSQ),
    "deerma-humidifier-mjjsq": partial(
        AirHumidifierMjjsq, model=MODEL_HUMIDIFIER_MJJSQ
    ),
    "deerma-humidifier-jsq1": partial(AirHumidifierMjjsq, model=MODEL_HUMIDIFIER_JSQ1),
    "yunmi-waterpuri-v2": WaterPurifier,
    "yunmi.waterpuri.lx9": WaterPurifierYunmi,
    "yunmi.waterpuri.lx11": WaterPurifierYunmi,
    "philips-light-bulb": PhilipsBulb,  # cannot be discovered via mdns
    "philips-light-hbulb": PhilipsWhiteBulb,  # cannot be discovered via mdns
    "philips-light-candle": PhilipsBulb,  # cannot be discovered via mdns
    "philips-light-candle2": PhilipsBulb,  # cannot be discovered via mdns
    "philips-light-ceiling": Ceil,
    "philips-light-zyceiling": Ceil,
    "philips-light-sread1": PhilipsEyecare,  # name needs to be checked
    "philips-light-moonlight": PhilipsMoonlight,  # name needs to be checked
    "philips-light-rwread": PhilipsRwread,  # name needs to be checked
    "xiaomi-wifispeaker-v1": WifiSpeaker,  # name needs to be checked
    "xiaomi-repeater-v1": WifiRepeater,  # name needs to be checked
    "xiaomi-repeater-v3": WifiRepeater,  # name needs to be checked
    "chunmi-cooker-press1": Cooker,
    "chunmi-cooker-press2": Cooker,
    "chunmi-cooker-normal1": Cooker,
    "chunmi-cooker-normal2": Cooker,
    "chunmi-cooker-normal3": Cooker,
    "chunmi-cooker-normal4": Cooker,
    "chunmi-cooker-normal5": Cooker,
    "lumi-acpartner-v1": partial(AirConditioningCompanion, model=MODEL_ACPARTNER_V1),
    "lumi-acpartner-v2": partial(AirConditioningCompanion, model=MODEL_ACPARTNER_V2),
    "lumi-acpartner-v3": partial(AirConditioningCompanion, model=MODEL_ACPARTNER_V3),
    "lumi-acpartner-mcn02": partial(
        AirConditioningCompanionMcn02, model=MODEL_ACPARTNER_MCN02
    ),
    "lumi-camera-aq2": AqaraCamera,
    "yeelink-light-": Yeelight,
    "leshow-fan-ss4": FanLeshow,
    "zhimi-fan-v2": partial(Fan, model=MODEL_FAN_V2),
    "zhimi-fan-v3": partial(Fan, model=MODEL_FAN_V3),
    "zhimi-fan-sa1": partial(Fan, model=MODEL_FAN_SA1),
    "zhimi-fan-za1": partial(Fan, model=MODEL_FAN_ZA1),
    "zhimi-fan-za3": partial(Fan, model=MODEL_FAN_ZA3),
    "zhimi-fan-za4": partial(Fan, model=MODEL_FAN_ZA4),
    "dmaker-fan-1c": partial(FanMiot, model=MODEL_FAN_1C),
    "dmaker-fan-p5": partial(Fan, model=MODEL_FAN_P5),
    "dmaker-fan-p9": partial(FanMiot, model=MODEL_FAN_P9),
    "dmaker-fan-p10": partial(FanMiot, model=MODEL_FAN_P10),
    "dmaker-fan-p11": partial(FanMiot, model=MODEL_FAN_P11),
    "tinymu-toiletlid-v1": partial(Toiletlid, model=MODEL_TOILETLID_V1),
    "zhimi-airfresh-va2": partial(AirFresh, model=MODEL_AIRFRESH_VA2),
    "zhimi-airfresh-va4": partial(AirFresh, model=MODEL_AIRFRESH_VA4),
    "dmaker-airfresh-t2017": AirFreshT2017,
    "zhimi-airmonitor-v1": partial(AirQualityMonitor, model=MODEL_AIRQUALITYMONITOR_V1),
    "cgllc-airmonitor-b1": partial(AirQualityMonitor, model=MODEL_AIRQUALITYMONITOR_B1),
    "cgllc-airmonitor-s1": partial(AirQualityMonitor, model=MODEL_AIRQUALITYMONITOR_S1),
    "lumi-gateway-": Gateway,
    "viomi-vacuum-v7": ViomiVacuum,
    "viomi-vacuum-v8": ViomiVacuum,
    "zhimi.heater.za1": partial(Heater, model=MODEL_HEATER_ZA1),
    "zhimi.elecheater.ma1": partial(Heater, model=MODEL_HEATER_MA1),
}


def pretty_token(token):
    """Return a pretty string presentation for a token."""
    return codecs.encode(token, "hex").decode()


def get_addr_from_info(info):
    addrs = info.addresses
    if len(addrs) > 1:
        _LOGGER.warning(
            "More than single IP address in the advertisement, using the first one"
        )

    return str(ip_address(addrs[0]))


def other_package_info(info, desc):
    """Return information about another package supporting the device."""
    return "Found %s at %s, check %s" % (info.name, get_addr_from_info(info), desc)


def create_device(name: str, addr: str, device_cls: partial) -> Device:
    """Return a device object for a zeroconf entry."""
    _LOGGER.debug(
        "Found a supported '%s', using '%s' class", name, device_cls.func.__name__
    )

    dev = device_cls(ip=addr)
    m = dev.send_handshake()
    dev.token = m.checksum
    _LOGGER.info(
        "Found a supported '%s' at %s - token: %s",
        device_cls.func.__name__,
        addr,
        pretty_token(dev.token),
    )
    return dev


class Listener(zeroconf.ServiceListener):
    """mDNS listener creating Device objects based on detected devices."""

    def __init__(self):
        self.found_devices = {}  # type: Dict[str, Device]

    def check_and_create_device(self, info, addr) -> Optional[Device]:
        """Create a corresponding :class:`Device` implementation for a given info and
        address.."""
        name = info.name
        for identifier, v in DEVICE_MAP.items():
            if name.startswith(identifier):
                if inspect.isclass(v):
                    return create_device(name, addr, partial(v))
                elif isinstance(v, partial) and inspect.isclass(v.func):
                    return create_device(name, addr, v)
                elif callable(v):
                    dev = Device(ip=addr)
                    _LOGGER.info(
                        "%s: token: %s",
                        v(info),
                        pretty_token(dev.send_handshake().checksum),
                    )
                    return None
        _LOGGER.warning(
            "Found unsupported device %s at %s, " "please report to developers",
            name,
            addr,
        )
        return None

    def add_service(self, zeroconf: "zeroconf.Zeroconf", type_: str, name: str) -> None:
        """Callback for discovery responses."""
        info = zeroconf.get_service_info(type_, name)
        addr = get_addr_from_info(info)

        if addr not in self.found_devices:
            dev = self.check_and_create_device(info, addr)
            if dev is not None:
                self.found_devices[addr] = dev

    def update_service(self, zc: "zeroconf.Zeroconf", type_: str, name: str) -> None:
        """Callback for state updates, which we ignore for now."""


class Discovery:
    """mDNS discoverer for miIO based devices (_miio._udp.local).

    Calling :func:`discover_mdns` will cause this to subscribe for updates on
    ``_miio._udp.local`` until any key is pressed, after which a dict of detected
    devices is returned.
    """

    @staticmethod
    def discover_mdns(*, timeout=5) -> Dict[str, Device]:
        """Discover devices with mdns until any keyboard input."""
        _LOGGER.info("Discovering devices with mDNS for %s seconds...", timeout)

        listener = Listener()
        browser = zeroconf.ServiceBrowser(
            zeroconf.Zeroconf(), "_miio._udp.local.", listener
        )

        time.sleep(timeout)
        browser.cancel()

        return listener.found_devices



import enum
import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_AIRFRESH_VA2 = "zhimi.airfresh.va2"
MODEL_AIRFRESH_VA4 = "zhimi.airfresh.va4"

AVAILABLE_PROPERTIES_COMMON = [
    "power",
    "temp_dec",
    "aqi",
    "average_aqi",
    "co2",
    "buzzer",
    "child_lock",
    "humidity",
    "led_level",
    "mode",
    "motor1_speed",
    "use_time",
    "ntcT",
    "app_extra",
    "f1_hour_used",
    "filter_life",
    "f_hour",
    "favorite_level",
    "led",
]

AVAILABLE_PROPERTIES = {
    MODEL_AIRFRESH_VA2: AVAILABLE_PROPERTIES_COMMON,
    MODEL_AIRFRESH_VA4: AVAILABLE_PROPERTIES_COMMON + ["ptc_state"],
}


class AirFreshException(DeviceException):
    pass


class OperationMode(enum.Enum):
    # Supported modes of the Air Fresh VA2 (zhimi.airfresh.va2)
    Auto = "auto"
    Silent = "silent"
    Interval = "interval"
    Low = "low"
    Middle = "middle"
    Strong = "strong"


class LedBrightness(enum.Enum):
    Bright = 0
    Dim = 1
    Off = 2


class AirFreshStatus(DeviceStatus):
    """Container for status reports from the air fresh."""

    def __init__(self, data: Dict[str, Any], model: str) -> None:
        """
        Response of a Air Fresh VA4 (zhimi.airfresh.va4):

        {
            'power': 'on',
            'temp_dec': 28.5,
            'aqi': 1,
            'average_aqi': 1,
            'co2': 1081,
            'buzzer': 'off',
            'child_lock': 'off',
            'humidity': 40,
            'led_level': 1,
            'mode': 'silent',
            'motor1_speed': 400,
            'use_time': 510000,
            'ntcT': 33.53,
            'app_extra': None,
            'f1_hour_used': 141,
            'filter_life': None,
            'f_hour': None,
            'favorite_level': None,
            'led': None,
            'ptc_state': 'off',
        }
        """

        self.data = data
        self.model = model

    @property
    def power(self) -> str:
        """Power state."""
        return self.data["power"]

    @property
    def is_on(self) -> bool:
        """Return True if device is on."""
        return self.power == "on"

    @property
    def aqi(self) -> int:
        """Air quality index."""
        return self.data["aqi"]

    @property
    def average_aqi(self) -> int:
        """Average of the air quality index."""
        return self.data["average_aqi"]

    @property
    def co2(self) -> int:
        """Carbon dioxide."""
        return self.data["co2"]

    @property
    def humidity(self) -> int:
        """Current humidity."""
        return self.data["humidity"]

    @property
    def ptc(self) -> Optional[bool]:
        """Return True if PTC is on."""
        if self.data["ptc_state"] is not None:
            return self.data["ptc_state"] == "on"

        return None

    @property
    def temperature(self) -> Optional[float]:
        """Current temperature, if available."""
        if self.data["temp_dec"] is not None:
            if self.model == MODEL_AIRFRESH_VA4:
                return self.data["temp_dec"]
            else:
                return self.data["temp_dec"] / 10.0

        return None

    @property
    def ntc_temperature(self) -> Optional[float]:
        """Current ntc temperature, if available."""
        if self.data["ntcT"] is not None:
            return self.data["ntcT"]

        return None

    @property
    def mode(self) -> OperationMode:
        """Current operation mode."""
        return OperationMode(self.data["mode"])

    @property
    def led(self) -> bool:
        """Return True if LED is on."""
        return self.data["led"] == "on"

    @property
    def led_brightness(self) -> Optional[LedBrightness]:
        """Brightness of the LED."""
        if self.data["led_level"] is not None:
            try:
                return LedBrightness(self.data["led_level"])
            except ValueError:
                _LOGGER.error(
                    "Unsupported LED brightness discarded: %s", self.data["led_level"]
                )
                return None

        return None

    @property
    def buzzer(self) -> Optional[bool]:
        """Return True if buzzer is on."""
        if self.data["buzzer"] is not None:
            return self.data["buzzer"] == "on"

        return None

    @property
    def child_lock(self) -> bool:
        """Return True if child lock is on."""
        return self.data["child_lock"] == "on"

    @property
    def filter_life_remaining(self) -> int:
        """Time until the filter should be changed."""
        return self.data["filter_life"]

    @property
    def filter_hours_used(self) -> int:
        """How long the filter has been in use."""
        return self.data["f1_hour_used"]

    @property
    def use_time(self) -> int:
        """How long the device has been active in seconds."""
        return self.data["use_time"]

    @property
    def motor_speed(self) -> int:
        """Speed of the motor."""
        return self.data["motor1_speed"]

    @property
    def extra_features(self) -> Optional[int]:
        return self.data["app_extra"]


class AirFresh(Device):
    """Main class representing the air fresh."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_AIRFRESH_VA2,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_AIRFRESH_VA2

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Heater (PTC): {result.ptc}\n"
            "AQI: {result.aqi} μg/m³\n"
            "Average AQI: {result.average_aqi} μg/m³\n"
            "Temperature: {result.temperature} °C\n"
            "NTC temperature: {result.ntc_temperature} °C\n"
            "Humidity: {result.humidity} %\n"
            "CO2: {result.co2} %\n"
            "Mode: {result.mode.value}\n"
            "LED: {result.led}\n"
            "LED brightness: {result.led_brightness}\n"
            "Buzzer: {result.buzzer}\n"
            "Child lock: {result.child_lock}\n"
            "Filter life remaining: {result.filter_life_remaining} %\n"
            "Filter hours used: {result.filter_hours_used}\n"
            "Use time: {result.use_time} s\n"
            "Motor speed: {result.motor_speed} rpm\n",
        )
    )
    def status(self) -> AirFreshStatus:
        """Retrieve properties."""

        properties = AVAILABLE_PROPERTIES[self.model]
        values = self.get_properties(properties, max_properties=15)

        return AirFreshStatus(
            defaultdict(lambda: None, zip(properties, values)), self.model
        )

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", ["on"])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", ["off"])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        return self.send("set_mode", [mode.value])

    @command(
        click.argument("led", type=bool),
        default_output=format_output(
            lambda led: "Turning on LED" if led else "Turning off LED"
        ),
    )
    def set_led(self, led: bool):
        """Turn led on/off."""
        if led:
            return self.send("set_led", ["on"])
        else:
            return self.send("set_led", ["off"])

    @command(
        click.argument("brightness", type=EnumType(LedBrightness)),
        default_output=format_output("Setting LED brightness to {brightness}"),
    )
    def set_led_brightness(self, brightness: LedBrightness):
        """Set led brightness."""
        return self.send("set_led_level", [brightness.value])

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set buzzer on/off."""
        if buzzer:
            return self.send("set_buzzer", ["on"])
        else:
            return self.send("set_buzzer", ["off"])

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        if lock:
            return self.send("set_child_lock", ["on"])
        else:
            return self.send("set_child_lock", ["off"])

    @command(
        click.argument("value", type=int),
        default_output=format_output("Setting extra to {value}"),
    )
    def set_extra_features(self, value: int):
        """Storage register to enable extra features at the app."""
        if value < 0:
            raise AirFreshException("Invalid app extra value: %s" % value)

        return self.send("set_app_extra", [value])

    @command(default_output=format_output("Resetting filter"))
    def reset_filter(self):
        """Resets filter hours used and remaining life."""
        return self.send("reset_filter1")

    @command(
        click.argument("ptc", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on PTC" if buzzer else "Turning off PTC"
        ),
    )
    def set_ptc(self, ptc: bool):
        """Set PTC on/off."""
        if ptc:
            return self.send("set_ptc_state", ["on"])
        else:
            return self.send("set_ptc_state", ["off"])


class AirFreshVA4(AirFresh):
    """Main class representing the air fresh va4."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
    ) -> None:
        super().__init__(
            ip, token, start_id, debug, lazy_discover, model=MODEL_AIRFRESH_VA4
        )



from unittest import TestCase

import pytest

from miio import FanLeshow
from miio.fan_leshow import (
    MODEL_FAN_LESHOW_SS4,
    FanLeshowException,
    FanLeshowStatus,
    OperationMode,
)

from .dummies import DummyDevice


class DummyFanLeshow(DummyDevice, FanLeshow):
    def __init__(self, *args, **kwargs):
        self.model = MODEL_FAN_LESHOW_SS4
        self.state = {
            "power": 1,
            "mode": 2,
            "blow": 100,
            "timer": 0,
            "sound": 1,
            "yaw": 0,
            "fault": 0,
        }
        self.return_values = {
            "get_prop": self._get_state,
            "set_power": lambda x: self._set_state("power", x),
            "set_mode": lambda x: self._set_state("mode", x),
            "set_blow": lambda x: self._set_state("blow", x),
            "set_timer": lambda x: self._set_state("timer", x),
            "set_sound": lambda x: self._set_state("sound", x),
            "set_yaw": lambda x: self._set_state("yaw", x),
        }
        super().__init__(args, kwargs)


@pytest.fixture(scope="class")
def fanleshow(request):
    request.cls.device = DummyFanLeshow()
    # TODO add ability to test on a real device


@pytest.mark.usefixtures("fanleshow")
class TestFanLeshow(TestCase):
    def is_on(self):
        return self.device.status().is_on

    def state(self):
        return self.device.status()

    def test_on(self):
        self.device.off()  # ensure off
        assert self.is_on() is False

        self.device.on()
        assert self.is_on() is True

    def test_off(self):
        self.device.on()  # ensure on
        assert self.is_on() is True

        self.device.off()
        assert self.is_on() is False

    def test_status(self):
        self.device._reset_state()

        assert repr(self.state()) == repr(FanLeshowStatus(self.device.start_state))

        assert self.is_on() is True
        assert self.state().mode == OperationMode(self.device.start_state["mode"])
        assert self.state().speed == self.device.start_state["blow"]
        assert self.state().buzzer is (self.device.start_state["sound"] == 1)
        assert self.state().oscillate is (self.device.start_state["yaw"] == 1)
        assert self.state().delay_off_countdown == self.device.start_state["timer"]
        assert self.state().error_detected is (self.device.start_state["fault"] == 1)

    def test_set_speed(self):
        def speed():
            return self.device.status().speed

        self.device.set_speed(0)
        assert speed() == 0
        self.device.set_speed(1)
        assert speed() == 1
        self.device.set_speed(100)
        assert speed() == 100

        with pytest.raises(FanLeshowException):
            self.device.set_speed(-1)

        with pytest.raises(FanLeshowException):
            self.device.set_speed(101)

    def test_set_oscillate(self):
        def oscillate():
            return self.device.status().oscillate

        self.device.set_oscillate(True)
        assert oscillate() is True

        self.device.set_oscillate(False)
        assert oscillate() is False

    def test_set_buzzer(self):
        def buzzer():
            return self.device.status().buzzer

        self.device.set_buzzer(True)
        assert buzzer() is True

        self.device.set_buzzer(False)
        assert buzzer() is False

    def test_delay_off(self):
        def delay_off_countdown():
            return self.device.status().delay_off_countdown

        self.device.delay_off(100)
        assert delay_off_countdown() == 100
        self.device.delay_off(200)
        assert delay_off_countdown() == 200
        self.device.delay_off(0)
        assert delay_off_countdown() == 0

        with pytest.raises(FanLeshowException):
            self.device.delay_off(-1)

        with pytest.raises(FanLeshowException):
            self.device.delay_off(541)



import enum
import logging
from collections import defaultdict
from typing import Any, Dict, Optional

import click

from .click_common import EnumType, command, format_output
from .device import Device, DeviceStatus
from .exceptions import DeviceException

_LOGGER = logging.getLogger(__name__)

MODEL_AIRFRESH_A1 = "dmaker.airfresh.a1"
MODEL_AIRFRESH_T2017 = "dmaker.airfresh.t2017"

AVAILABLE_PROPERTIES_COMMON = [
    "power",
    "mode",
    "pm25",
    "co2",
    "temperature_outside",
    "favourite_speed",
    "control_speed",
    "ptc_on",
    "ptc_status",
    "child_lock",
    "sound",
    "display",
]

AVAILABLE_PROPERTIES = {
    MODEL_AIRFRESH_T2017: AVAILABLE_PROPERTIES_COMMON
    + [
        "filter_intermediate",
        "filter_inter_day",
        "filter_efficient",
        "filter_effi_day",
        "ptc_level",
        "screen_direction",
    ],
    MODEL_AIRFRESH_A1: AVAILABLE_PROPERTIES_COMMON
    + [
        "filter_rate",
        "filter_day",
    ],
}


class AirFreshException(DeviceException):
    pass


class OperationMode(enum.Enum):
    Off = "off"
    Auto = "auto"
    Sleep = "sleep"
    Favorite = "favourite"


class PtcLevel(enum.Enum):
    Low = "low"
    Medium = "medium"
    High = "high"


class DisplayOrientation(enum.Enum):
    Portrait = "forward"
    LandscapeLeft = "left"
    LandscapeRight = "right"


class AirFreshStatus(DeviceStatus):
    """Container for status reports from the air fresh t2017."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """
        Response of a Air Fresh A1 (dmaker.airfresh.a1):
        {
            'power': True,
            'mode': 'auto',
            'pm25': 2,
            'co2': 554,
            'temperature_outside': 12,
            'favourite_speed': 150,
            'control_speed': 60,
            'filter_rate': 45,
            'filter_day': 81,
            'ptc_on': False,
            'ptc_status': False,
            'child_lock': False,
            'sound': False,
            'display': False,
        }

        Response of a Air Fresh T2017 (dmaker.airfresh.t2017):

        {
            'power': True,
            'mode': 'favourite',
            'pm25': 1,
            'co2': 550,
            'temperature_outside': 24,
            'favourite_speed': 241,
            'control_speed': 241,
            'filter_intermediate': 100,
            'filter_inter_day': 90,
            'filter_efficient': 100,
            'filter_effi_day': 180,
            'ptc_on': False,
            'ptc_level': 'low',
            'ptc_status': False,
            'child_lock': False,
            'sound': True,
            'display': False,
            'screen_direction': 'forward',
        }
        """

        self.data = data

    @property
    def power(self) -> str:
        """Power state."""
        return "on" if self.data["power"] else "off"

    @property
    def is_on(self) -> bool:
        """Return True if device is on."""
        return self.data["power"]

    @property
    def mode(self) -> OperationMode:
        """Current operation mode."""
        return OperationMode(self.data["mode"])

    @property
    def pm25(self) -> int:
        """Fine particulate patter (PM2.5)."""
        return self.data["pm25"]

    @property
    def co2(self) -> int:
        """Carbon dioxide."""
        return self.data["co2"]

    @property
    def temperature(self) -> int:
        """Current temperature in degree celsions."""
        return self.data["temperature_outside"]

    @property
    def favorite_speed(self) -> int:
        """Favorite speed."""
        return self.data["favourite_speed"]

    @property
    def control_speed(self) -> int:
        """Control speed."""
        return self.data["control_speed"]

    @property
    def dust_filter_life_remaining(self) -> Optional[int]:
        """Remaining dust filter life in percent."""
        return self.data.get("filter_intermediate", self.data.get("filter_rate"))

    @property
    def dust_filter_life_remaining_days(self) -> Optional[int]:
        """Remaining dust filter life in days."""
        return self.data.get("filter_inter_day", self.data.get("filter_day"))

    @property
    def upper_filter_life_remaining(self) -> Optional[int]:
        """Remaining upper filter life in percent."""
        return self.data.get("filter_efficient")

    @property
    def upper_filter_life_remaining_days(self) -> Optional[int]:
        """Remaining upper filter life in days."""
        return self.data.get("filter_effi_day")

    @property
    def ptc(self) -> bool:
        """Return True if PTC is on."""
        return self.data["ptc_on"]

    @property
    def ptc_level(self) -> Optional[PtcLevel]:
        """PTC level."""
        try:
            return PtcLevel(self.data["ptc_level"])
        except (KeyError, ValueError):
            return None

    @property
    def ptc_status(self) -> bool:
        """Return true if PTC status is on."""
        return self.data["ptc_status"]

    @property
    def child_lock(self) -> bool:
        """Return True if child lock is on."""
        return self.data["child_lock"]

    @property
    def buzzer(self) -> bool:
        """Return True if sound is on."""
        return self.data["sound"]

    @property
    def display(self) -> bool:
        """Return True if the display is on."""
        return self.data["display"]

    @property
    def display_orientation(self) -> Optional[DisplayOrientation]:
        """Display orientation."""
        try:
            return DisplayOrientation(self.data["screen_direction"])
        except (KeyError, ValueError):
            return None


class AirFreshA1(Device):
    """Main class representing the air fresh a1."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_AIRFRESH_A1,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_AIRFRESH_A1

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode}\n"
            "PM2.5: {result.pm25}\n"
            "CO2: {result.co2}\n"
            "Temperature: {result.temperature}\n"
            "Favorite speed: {result.favorite_speed}\n"
            "Control speed: {result.control_speed}\n"
            "Dust filter life: {result.dust_filter_life_remaining} %, "
            "{result.dust_filter_life_remaining_days} days\n"
            "PTC: {result.ptc}\n"
            "PTC status: {result.ptc_status}\n"
            "Child lock: {result.child_lock}\n"
            "Buzzer: {result.buzzer}\n"
            "Display: {result.display}\n",
        )
    )
    def status(self) -> AirFreshStatus:
        """Retrieve properties."""

        properties = AVAILABLE_PROPERTIES[self.model]
        values = self.get_properties(properties, max_properties=15)

        return AirFreshStatus(defaultdict(lambda: None, zip(properties, values)))

    @command(default_output=format_output("Powering on"))
    def on(self):
        """Power on."""
        return self.send("set_power", [True])

    @command(default_output=format_output("Powering off"))
    def off(self):
        """Power off."""
        return self.send("set_power", [False])

    @command(
        click.argument("mode", type=EnumType(OperationMode)),
        default_output=format_output("Setting mode to '{mode.value}'"),
    )
    def set_mode(self, mode: OperationMode):
        """Set mode."""
        return self.send("set_mode", [mode.value])

    @command(
        click.argument("display", type=bool),
        default_output=format_output(
            lambda led: "Turning on display" if led else "Turning off display"
        ),
    )
    def set_display(self, display: bool):
        """Turn led on/off."""
        return self.send("set_display", [display])

    @command(
        click.argument("ptc", type=bool),
        default_output=format_output(
            lambda ptc: "Turning on ptc" if ptc else "Turning off ptc"
        ),
    )
    def set_ptc(self, ptc: bool):
        """Turn ptc on/off."""
        return self.send("set_ptc_on", [ptc])

    @command(
        click.argument("buzzer", type=bool),
        default_output=format_output(
            lambda buzzer: "Turning on buzzer" if buzzer else "Turning off buzzer"
        ),
    )
    def set_buzzer(self, buzzer: bool):
        """Set sound on/off."""
        return self.send("set_sound", [buzzer])

    @command(
        click.argument("lock", type=bool),
        default_output=format_output(
            lambda lock: "Turning on child lock" if lock else "Turning off child lock"
        ),
    )
    def set_child_lock(self, lock: bool):
        """Set child lock on/off."""
        return self.send("set_child_lock", [lock])

    @command(default_output=format_output("Resetting dust filter"))
    def reset_dust_filter(self):
        """Resets filter lifetime of the dust filter."""
        return self.send("set_filter_rate", [100])

    @command(
        click.argument("speed", type=int),
        default_output=format_output("Setting favorite speed to {speed}"),
    )
    def set_favorite_speed(self, speed: int):
        """Sets the fan speed in favorite mode."""
        if speed < 0 or speed > 150:
            raise AirFreshException("Invalid favorite speed: %s" % speed)

        return self.send("set_favourite_speed", [speed])

    @command()
    def set_ptc_timer(self):
        """
        value = time.index + '-' +
            time.hexSum + '-' +
            time.startTime + '-' +
            time.ptcTimer.endTime + '-' +
            time.level + '-' +
            time.status;
        return self.send("set_ptc_timer", [value])
        """
        raise NotImplementedError()

    @command()
    def get_ptc_timer(self):
        """Returns a list of PTC timers.

        Response unknown.
        """
        return self.send("get_ptc_timer")

    @command()
    def get_timer(self):
        """Response unknown."""
        return self.send("get_timer")


class AirFreshT2017(AirFreshA1):
    """Main class representing the air fresh t2017."""

    def __init__(
        self,
        ip: str = None,
        token: str = None,
        start_id: int = 0,
        debug: int = 0,
        lazy_discover: bool = True,
        model: str = MODEL_AIRFRESH_T2017,
    ) -> None:
        super().__init__(ip, token, start_id, debug, lazy_discover)

        if model in AVAILABLE_PROPERTIES:
            self.model = model
        else:
            self.model = MODEL_AIRFRESH_T2017

    @command(
        default_output=format_output(
            "",
            "Power: {result.power}\n"
            "Mode: {result.mode}\n"
            "PM2.5: {result.pm25}\n"
            "CO2: {result.co2}\n"
            "Temperature: {result.temperature}\n"
            "Favorite speed: {result.favorite_speed}\n"
            "Control speed: {result.control_speed}\n"
            "Dust filter life: {result.dust_filter_life_remaining} %, "
            "{result.dust_filter_life_remaining_days} days\n"
            "Upper filter life remaining: {result.upper_filter_life_remaining} %, "
            "{result.upper_filter_life_remaining_days} days\n"
            "PTC: {result.ptc}\n"
            "PTC level: {result.ptc_level}\n"
            "PTC status: {result.ptc_status}\n"
            "Child lock: {result.child_lock}\n"
            "Buzzer: {result.buzzer}\n"
            "Display: {result.display}\n"
            "Display orientation: {result.display_orientation}\n",
        )
    )
    def status(self) -> AirFreshStatus:
        """Retrieve properties."""

        return super().status()

    @command(
        click.argument("speed", type=int),
        default_output=format_output("Setting favorite speed to {speed}"),
    )
    def set_favorite_speed(self, speed: int):
        """Sets the fan speed in favorite mode."""
        if speed < 60 or speed > 300:
            raise AirFreshException("Invalid favorite speed: %s" % speed)

        return self.send("set_favourite_speed", [speed])

    @command(default_output=format_output("Resetting dust filter"))
    def reset_dust_filter(self):
        """Resets filter lifetime of the dust filter."""
        return self.send("set_filter_reset", ["intermediate"])

    @command(default_output=format_output("Resetting upper filter"))
    def reset_upper_filter(self):
        """Resets filter lifetime of the upper filter."""
        return self.send("set_filter_reset", ["efficient"])

    @command(
        click.argument("orientation", type=EnumType(DisplayOrientation)),
        default_output=format_output("Setting orientation to '{orientation.value}'"),
    )
    def set_display_orientation(self, orientation: DisplayOrientation):
        """Set display orientation."""
        return self.send("set_screen_direction", [orientation.value])

    @command(
        click.argument("level", type=EnumType(PtcLevel)),
        default_output=format_output("Setting ptc level to '{level.value}'"),
    )
    def set_ptc_level(self, level: PtcLevel):
        """Set PTC level."""
        return self.send("set_ptc_level", [level.value])
